# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import namedtuple
import torch


VarParams = namedtuple('VarParams', ['var', 'lr_t', 'm_t', 'beta1_broad', 'grad', 'epsilon', 'v_t'])


def _output_m_compute(m, beta1_broad, grad):
    """
    _output_m_compute
    do compute m_t = m + (beta1 - 1) * (m - grad)
    """
    input_dtype = m.dtype

    sneg_one = torch.ones((1), dtype=input_dtype) * -1
    sneg_one = sneg_one.to(beta1_broad.device)

    # `formula; beta1 -1`
    vsub_beta1_1 = torch.add(beta1_broad, sneg_one)

    # `formula; m - grad`
    vsub_m_grad = torch.sub(m, grad)

    # `formula; (beta1 - 1) * (m - grad)`
    vmul_m = torch.mul(vsub_beta1_1, vsub_m_grad)

    # `formula; m_t = m + (beta1 - 1) * (m - grad)`
    m_t = torch.add(m, vmul_m)

    return m_t


def _output_v_compute(v, beta2, grad):
    """
    _output_v_compute
    do compute v_t = v + (1 - beta2)*(grad*grad -v)
    """
    input_dtype = v.dtype

    sneg_one = torch.ones((1), dtype=input_dtype) * -1

    # `formula; broadcast beta2 to vector`
    beta2_tensor = torch.tensor(beta2, dtype=input_dtype)
    beta2_broad = beta2_tensor.expand_as(v)

    # `formula; beta2 - 1`
    vsub_beta2_1 = torch.add(beta2_broad, sneg_one)
    vsub_beta2_1 = vsub_beta2_1.to(v.device)

    # `formula; grad * grad`
    vmul_grad_grad = torch.mul(grad, grad)

    # `formula; (v - grad*grad)`
    vsub_v_grad = torch.sub(v, vmul_grad_grad)

    # `formula; (beta2 -1) * (v - grad * grad)`
    vmul_grad = torch.mul(vsub_beta2_1, vsub_v_grad)

    # `formula; v_t = v + (beta2 - 1) * (v - grad * grad)`
    v_t = torch.add(v, vmul_grad)

    return v_t


def _inner_lr_compute(lr, beta2_power, beta1_power, compute_shape_tensor):
    """
    _inner_lr_compute
    `formula; lr_t = learning_rate * (sqrt(1-beta2_power)) / (1 - beta1_power)`
    """

    input_dtype = compute_shape_tensor.dtype

    s_one = torch.ones((1), dtype=input_dtype)

    s_neg_one = torch.ones((1), dtype=input_dtype) * -1

    # `formula; (1 - beta2_power)`
    v_neg_beta2_power = torch.mul(beta2_power, s_neg_one)
    v_add_beta2_power = torch.add(v_neg_beta2_power, s_one)

    # `formula; sqrt(1 - beta2_power)`
    v_sqrt_beta2_power = torch.sqrt(v_add_beta2_power)

    # `formula; (1 - beta1_power)`
    v_neg_beta1_power = torch.mul(beta1_power, s_neg_one)
    v_add_beta1_power = torch.add(v_neg_beta1_power, s_one)

    # `formula; learning_rate * (sqrt(1-beta2_power)`
    res = torch.mul(lr, v_sqrt_beta2_power)

    # `formula; learning_rate*(sqrt(1-beta2_power))/(1-beta1_power)`
    res = torch.div(res, v_add_beta1_power)
    return res.expand_as(compute_shape_tensor)


def _inner_eps_add_sqrt_vt_compute(epsilon, v_t):
    """
    (epsilon + sqrt(v_t) )
    """
    # `formula; sqrt(v_t)`
    sqrt_vt = torch.sqrt(v_t)

    # `formula; broadcast epsilon  to vector`
    input_dtype = v_t.dtype
    epsilon_tensor = torch.tensor(epsilon, dtype=input_dtype)
    epsilon_broad = epsilon_tensor.expand_as(v_t)
    epsilon_broad = epsilon_broad.to(sqrt_vt.device)

    # `formula; epsilon + sqrt(v_t)`
    v_add_sqrt_v = torch.add(sqrt_vt, epsilon_broad)

    return v_add_sqrt_v


def _output_var_t_compute_use_nesterov(varparams):
    """
    _output_var_t_compute_use_nesterov
    `formula; var_t = var - lr_t * (m_t * beta1 + (1 - beta1) * grad) / (epsilon + sqrt(v_t))`
    `formula; var_t = var - lr_t * (m_t * beta1 + (1 - beta1) * grad) / (epsilon + sqrt(v_t))`
    """
    var = varparams.var
    lr_t = varparams.lr_t
    m_t = varparams.m_t
    beta1_broad = varparams.beta1_broad
    grad = varparams.grad
    epsilon = varparams.epsilon
    v_t = varparams.v_t

    input_dtype = var.dtype

    s_one = torch.ones((1), dtype=input_dtype)

    s_neg_one = torch.ones((1), dtype=input_dtype) * -1

    # `formula; m_t * beta1`
    v_muls_mt_beta1 = torch.mul(m_t, beta1_broad)

    # `formula; 1 -beta1`
    v_neg_beta1 = torch.mul(beta1_broad, s_neg_one)
    vsub_1_beta1 = torch.add(v_neg_beta1, s_one)

    # `formula; (1-beta1)* grad`
    v_mul_grad = torch.mul(vsub_1_beta1, grad)

    # `formula; (m_t*beta1 + (1 - beta1)*grad)`
    v_div_left = torch.add(v_muls_mt_beta1, v_mul_grad)

    # `formula; lr_t * (m_t*beta1 + (1 - beta1) * grad)`
    # broadcast lr_t to vector

    lrt_broad = lr_t.expand_as(var)
    v_mul_left = torch.mul(lrt_broad, v_div_left)

    # `formula; (epsilon + sqrt(v_t))`
    v_add_sqrt_v = _inner_eps_add_sqrt_vt_compute(epsilon, v_t)

    # `formula; lr_t * (m_t*beta1 + (1-beta1)*grad / (epsilon + sqrt(v_t))`
    v_div_res = torch.div(v_mul_left, v_add_sqrt_v)

    # `formula; var - lr_t * (m_t*beta1 + (1-beta1)*grad) / (epsilon + sqrt(v_t))`
    v_t = torch.sub(var, v_div_res)

    return v_t


def _output_var_t_compute(var, lr_t, m_t, epsilon, v_t):
    """
    _output_var_t_compute
    `var_t = var - lr_t * m_t / (epsilon + sqrt(v_t))`
    """
    # `formula; lr_t * m_t`
    lr_t = lr_t.to(m_t.device)
    v_mul_left = torch.mul(lr_t, m_t)

    # `formula; (epsilon + sqrt(v_t))`
    v_add_sqrt_v = _inner_eps_add_sqrt_vt_compute(epsilon, v_t)

    # `formula; lr_t * m_t /(epsilon + sqrt(v_t))`
    v_div_res = torch.div(v_mul_left, v_add_sqrt_v)

    # `formula; var - lr_t * m_t / (epsilon + sqrt(v_t))`
    v_t = torch.sub(var, v_div_res)

    return v_t


def npu_apply_adam(beta1_power, beta2_power, lr, beta1, beta2, epsilon, grad, use_locking, use_nesterov, out):
    var, m, v = out
    input_dtype = m.dtype
    beta1_tensor = torch.tensor(beta1, dtype=input_dtype).to(m.device)
    beta1_broad = beta1_tensor.expand_as(m)
    m_t = _output_m_compute(m, beta1_broad, grad)
    v_t = _output_v_compute(v, beta2, grad)
    lr_t = _inner_lr_compute(lr, beta2_power, beta1_power, grad)
    if use_nesterov:
        var_params = VarParams(var, lr_t, m_t, beta1_broad, grad, epsilon, v_t)
        var_t = _output_var_t_compute_use_nesterov(var_params)
    else:
        var_t = _output_var_t_compute(var, lr_t, m_t, epsilon, v_t)
    return var_t, m_t, v_t



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import defaultdict
from functools import wraps

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.log import logger

# 记录工具函数递归的深度
recursion_depth = defaultdict(int)


def recursion_depth_decorator(func_info, max_depth=Const.MAX_DEPTH):
    """装饰一个函数，当函数递归调用超过限制时，抛出异常并打印函数信息。"""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func_id = id(func)
            recursion_depth[func_id] += 1
            if recursion_depth[func_id] > max_depth:
                msg = f"call {func_info} exceeds the recursion limit."
                logger.error_log_with_exp(
                    msg,
                    MsprobeException(
                        MsprobeException.RECURSION_LIMIT_ERROR, msg
                    ),
                )
            try:
                result = func(*args, **kwargs)
            finally:
                recursion_depth[func_id] -= 1
            return result

        return wrapper

    return decorator



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.const import Const, FileCheckConst, MsgConst
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.common.file_utils import FileChecker, load_json
from msprobe.core.common.utils import get_real_step_or_rank, check_init_step, ThreadSafe
from msprobe.core.dump.common_config import CommonConfig


class BasePrecisionDebugger:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with ThreadSafe():
                if not cls._instance:
                    cls._instance = super(BasePrecisionDebugger, cls).__new__(cls)
                    cls._instance.config = None
                    cls._instance.initialized = False
                    cls.service = None
                    cls.first_start = False
        return cls._instance

    def __init__(
            self,
            config_path=None,
            task=None,
            dump_path=None,
            level=None,
            step=None
    ):
        if self.initialized:
            return
        self.initialized = True
        self._check_input_params(config_path, task, dump_path, level)
        self.common_config, self.task_config = self._parse_config_path(config_path, task)
        self.task = self.common_config.task
        if step is not None:
            self.common_config.step = get_real_step_or_rank(step, Const.STEP)

    @staticmethod
    def _check_input_params(config_path, task, dump_path, level):
        if not config_path:
            config_path = os.path.join(os.path.dirname(__file__), "../../../config.json")
        if config_path is not None:
            if not isinstance(config_path, str):
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR, f"config_path must be a string")
            file_checker = FileChecker(
                file_path=config_path, path_type=FileCheckConst.FILE, file_type=FileCheckConst.JSON_SUFFIX)
            file_checker.common_check()

        if task is not None and task not in Const.TASK_LIST:
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, f"task must be one of {Const.TASK_LIST}")

        if dump_path is not None:
            if not isinstance(dump_path, str):
                raise MsprobeException(
                    MsprobeException.INVALID_PARAM_ERROR, f"dump_path must be a string")

        if level is not None and level not in Const.LEVEL_LIST:
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, f"level must be one of {Const.LEVEL_LIST}")

    @staticmethod
    def _get_task_config(task, json_config):
        raise NotImplementedError("Subclass must implement _get_task_config")

    @classmethod
    @ThreadSafe.synchronized
    def set_init_step(cls, step):
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        check_init_step(step)
        instance.service.init_step = step
        instance.service.loop = 0

    @classmethod
    @ThreadSafe.synchronized
    def register_custom_api(cls, module, api, api_prefix=None):
        if not api_prefix:
            api_prefix = getattr(module, "__name__", "Custom")
        if not isinstance(api_prefix, str):
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, "api_prefix must be string")
        if not hasattr(module, api):
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, f"module {str(module)} does not have {api}")
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        instance.service.register_custom_api(module, api, api_prefix)

    @classmethod
    @ThreadSafe.synchronized
    def restore_custom_api(cls, module, api):
        if not hasattr(module, api):
            raise MsprobeException(
                MsprobeException.INVALID_PARAM_ERROR, f"module {str(module)} does not have {api}")
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        instance.service.restore_custom_api(module, api)

    @classmethod
    def _get_instance(cls):
        instance = cls._instance
        if not instance:
            raise Exception(MsgConst.NOT_CREATED_INSTANCE)
        return instance

    def _parse_config_path(self, json_file_path, task):
        if not json_file_path:
            json_file_path = os.path.join(os.path.dirname(__file__), "../../../config.json")
        json_config = load_json(json_file_path)
        common_config = CommonConfig(json_config)
        if task:
            task_config = self._get_task_config(task, json_config)
        else:
            if not common_config.task:
                common_config.task = Const.STATISTICS
            task_config = self._get_task_config(common_config.task, json_config)
        return common_config, task_config



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
convert format from NC1HWC0 to NCHW.
"""
import numpy as np
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


def convert(shape_from: list, shape_to: list, array: any) -> any:
    """
    Convert the data format from NC1HWC0 to NCHW
    :param shape_from: the shape before convert
    :param shape_to: the shape after convert
    :param array: the one-dimensional array
    :return: the data array of NCHW shape
    """
    _ = shape_to
    if len(shape_from) < 5:
        log.print_error_log("length of shape of NC1HWC0 is less than 5, please check.")
        raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)
    n_from = shape_from[0]
    c1_from = shape_from[1]
    h_from = shape_from[2]
    w_from = shape_from[3]
    c0_from = shape_from[4]

    array_shape = array.reshape(n_from, c1_from, h_from, w_from, c0_from)
    tmp_input_tensor = np.transpose(array_shape, axes=(0, 1, 4, 2, 3))
    tmp_input_tensor = tmp_input_tensor.reshape((n_from, c1_from * c0_from, h_from, w_from))
    return tmp_input_tensor



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
import json
from typing import List

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.log import logger
from msprobe.core.compare.indicator_analysis.utils import CompareMode, ResultLevel, divide_result_df, IgnoreInfo
from msprobe.core.compare.indicator_analysis.algorithm import BaseAlgorithm, TENSOR_CHECKERS, STATISTICS_CHECKERS, \
    MD5_CHECKERS, STATISTICS_CHECKERS_PARALLEL_MERGE
from msprobe.core.compare.indicator_analysis.api_data import ApiData


class ApiIndicatorCalculator:
    RANK_SUFFIX_PATTERN = re.compile(r'_rank\d+$')

    def __init__(self, mode, parallel_merge=False):
        self.mode = mode
        self.parallel_merge = parallel_merge
        self.all_ignore_set = {'empty', 'empty_like', 'numpy', 'to', '__setitem__', 'empty_with_format',
                               'new_empty_strided', 'new_empty', 'empty_strided'}
        self.input_ignore_set = {'_reduce_scatter_base', '_all_gather_base', 'all_to_all_single', 'batch_isend_irecv'}
        self.algorithms: List[BaseAlgorithm] = []
        self._add_algorithm()

    @staticmethod
    def get_api_indicator_and_msg(api_data: ApiData):
        """
        基于all_data_lists（api或模块的所有参数的数据）得到一个api或模块的指标和异常信息

        indicator取所有参数最差的（error）
        err msg取所有参数汇总

        Return:
            精度比对指标（pass/warning/error）
        """
        all_data_lists = api_data.input_data + api_data.output_data
        final_indicator = ResultLevel.PASS
        for data_list in all_data_lists:
            indicator = api_data.get_data_by_header(CompareConst.RESULT, data_list)
            if isinstance(indicator, ResultLevel):
                api_data.set_data_by_header(CompareConst.RESULT, data_list, indicator.value)
                if indicator > final_indicator:
                    final_indicator = indicator
            err_msg = api_data.get_data_by_header(CompareConst.ERROR_MESSAGE, data_list)
            if isinstance(err_msg, list):
                api_data.set_data_by_header(CompareConst.ERROR_MESSAGE, data_list, json.dumps(err_msg))

        return final_indicator.value

    def get_api_ignore_info(self, api_data: ApiData):
        """
        api是否需要忽略判断规则的情况
        """
        if not api_data.input_data:
            return IgnoreInfo.NO_IGNORE
        npu_param_name = api_data.get_data_by_header(CompareConst.NPU_NAME, api_data.input_data[0])
        name_split = npu_param_name.split(Const.SEP)
        if len(name_split) < 2:
            return IgnoreInfo.NO_IGNORE
        api_name = self.RANK_SUFFIX_PATTERN.sub('', name_split[1]) if self.parallel_merge else name_split[1]
        if api_name in self.all_ignore_set:
            return IgnoreInfo.ALL_IGNORE
        elif api_name in self.input_ignore_set:
            return IgnoreInfo.INPUT_IGNORE

        return IgnoreInfo.NO_IGNORE

    def calculate(self, raw_data_list: List[List]):
        """
        计算入口
        """
        api_data = ApiData(self.mode, raw_data_list)

        ignore_info = self.get_api_ignore_info(api_data)

        self.execute_all(api_data, ignore_info)

        return self.get_api_indicator_and_msg(api_data)

    def add_algorithm(self, algorithm: BaseAlgorithm):
        if not isinstance(algorithm, BaseAlgorithm):
            msg = 'It must be an instance of a subclass of BaseAlgorithm.'
            logger.error(msg)
            raise TypeError(msg)
        self.algorithms.append(algorithm)

    def execute_all(self, api_data: ApiData, ignore_info: IgnoreInfo):
        for algorithm in self.algorithms:
            try:
                algorithm.run(api_data, ignore_info)
            except Exception as e:
                msg = f'Run algorithm failed.'
                logger.error(msg)
                raise RuntimeError(msg) from e

    def _add_algorithm(self):
        if self.mode == CompareMode.STATISTICS.value:
            checkers = STATISTICS_CHECKERS_PARALLEL_MERGE if self.parallel_merge else STATISTICS_CHECKERS
            for checker in checkers:
                self.add_algorithm(checker())
        elif self.mode == CompareMode.TENSOR.value:
            for checker in TENSOR_CHECKERS:
                self.add_algorithm(checker())
        elif self.mode == CompareMode.MD5.value:
            for checker in MD5_CHECKERS:
                self.add_algorithm(checker())


def calculate_excel_result_df(result_df, mode, chunk_size=1000):
    """
    仅适用于excel比对场景，得到表格每行数据的精度比对指标（pass/warning/error）

    Args:
        result_df: DataFrame数据结构，即转换成excel前的表单结构
        mode: 比对模式，分为 tensor 模式、统计量模式和 md5 模式
        chunk_size: 分块赋值参数，默认1000，把 result 分成小块，逐块赋值给 result_df，这样每次只占用小块内存，避免内存峰值过高
    """
    result_dict = divide_result_df(result_df)
    calculator = ApiIndicatorCalculator(mode)
    calculated_result_lists = []
    for data_lists in result_dict.values():
        calculator.calculate(data_lists)
        calculated_result_lists.extend(data_lists)

    head = CompareConst.HEAD_OF_COMPARE_MODE.get(mode)
    if not head:
        logger.error(f'Unable to obtain header based on compare mode: {mode}')
        raise RuntimeError()
    # 配置列映射关系：[(result_df的目标列名, result子列表的列索引)]
    try:
        cols_mapping = [
            (CompareConst.RESULT, head.index(CompareConst.RESULT)),
            (CompareConst.ERROR_MESSAGE, head.index(CompareConst.ERROR_MESSAGE))
        ]
    except ValueError as e:
        logger.error(f'The {CompareConst.RESULT} or {CompareConst.ERROR_MESSAGE} does not exist in the header: {e}')
        raise e

    total_rows = len(calculated_result_lists)

    # 分块逐批赋值，降低内存瞬时峰值
    for i in range(0, total_rows, chunk_size):
        end_idx = min(i + chunk_size, total_rows)
        current_result_chunk = calculated_result_lists[i:end_idx]

        for df_col_name, result_col_idx in cols_mapping:
            col_data = [sublist[result_col_idx] for sublist in current_result_chunk]
            df_col_idx = result_df.columns.get_loc(df_col_name)
            result_df.iloc[i:end_idx, df_col_idx] = col_data


def calculate_result(result, mode, parallel_merge=False):
    """
    得到一个api或模块的指标和异常信息

    Args:
        result: List[List]数据结构，每个list元素代表api或模块参数的具体信息
        mode: 比对模式，分为 tensor 模式、统计量模式和 md5 模式
        parallel_merge: 是否为不同切分策略图合并比对场景，默认False

    Return:
        精度比对指标（pass/warning/error）
    """
    calculator = ApiIndicatorCalculator(mode, parallel_merge)
    return calculator.calculate(result)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def matmul_backward(grad, self, other, mask):
    if len(mask) < 2:
        raise RuntimeError("Mask size at least 2")

    grad_self, grad_other = None, None
    dim_self = self.dim()
    dim_other = other.dim()

    size_grad = list(grad.size())
    size_self = list(self.size())
    size_other = list(other.size())

    if dim_self == 1 and dim_other == 1:
        grad_self = other.mul(grad) if mask[0] else grad_self
        grad_other = self.mul(grad) if mask[1] else grad_other
    elif dim_self == 2 and dim_other == 1:
        grad_self = grad.unsqueeze(1).mm(other.unsqueeze(0)) if mask[0] else grad_self
        grad_other = self.transpose(-1, -2).mm(grad.unsqueeze(1)).squeeze_(1) if mask[1] else grad_other
    elif dim_self == 1 and dim_other == 2:
        grad_self = grad.unsqueeze(0).mm(other.transpose(-1, -2)).squeeze_(0) if mask[0] else grad_self
        grad_other = self.unsqueeze(1).mm(grad.unsqueeze(0)) if mask[1] else grad_other
    elif dim_self >= 3 and (dim_other == 1 or dim_other == 2):
        if len(size_grad) < 1:
            raise RuntimeError("size_grad's length at least 1")
        view_size = 1 if dim_other == 1 else size_grad[-1]
        unfolded_grad = (grad.unsqueeze(-1) if dim_other == 1 else grad).contiguous().view(-1, view_size)
        if mask[0]:
            grad_self = unfolded_grad.mm(other.unsqueeze(0) if dim_other == 1 else other.transpose(-1, -2)) \
                .view(size_self)
        if mask[1]:
            if len(size_self) < 1:
                raise RuntimeError("size_self's length at least 1")
            unfolded_self = self.contiguous().view([-1, size_self[-1]])
            grad_other = unfolded_self.transpose(-1, -2).mm(unfolded_grad).view(size_other)
    elif (dim_self == 1 or dim_self == 2) and dim_other >= 3:
        if len(size_grad) < 2:
            raise RuntimeError("size_grad's length at least 2")
        view_size = 1 if dim_self == 1 else size_grad[-2]
        unfolded_grad_t = grad.view([-1, view_size]) \
            if dim_self == 1 else grad.transpose(-1, -2).contiguous().view([-1, view_size])
        if mask[0]:
            if len(size_other) < 2:
                raise RuntimeError("size_other's length at least 2")
            # create a 2D-matrix from other
            unfolded_other_t = \
                other.transpose(-1, -2).contiguous().view([-1, size_other[-2]]).transpose(-1, -2)
            grad_self = unfolded_other_t.mm(unfolded_grad_t).transpose(-1, -2).view(size_self)
        if mask[1]:
            size_other_t = size_other[:-2]
            size_other_t.extend(size_other[::-1][:2])
            grad_other = \
                unfolded_grad_t.mm(self.unsqueeze(0) if dim_self == 1 else self).view(size_other_t).transpose(-1, -2)
    else:
        grad_self = torch.matmul(grad, other.transpose(-1, -2)) if mask[0] else grad_self
        grad_other = torch.matmul(self.transpose(-1, -2), grad) if mask[1] else grad_other

    return grad_self.cpu(), grad_other.cpu()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import sys
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Optional

from tqdm import tqdm

from msprobe.core.common.const import Const, Data2DBConst
from msprobe.core.common.file_utils import (
    check_file_or_directory_path,
    create_directory,
    load_json,
    recursive_chmod,
    remove_path,
)
from msprobe.core.common.log import logger
from msprobe.core.dump.dump2db.db_utils import DumpDB


def validate_micro_step(micro_step):
    """
    校验 micro_step 值是否在有效范围内
    """
    if micro_step is not None:
        if type(micro_step) is not int:
            raise ValueError("Micro step must be an integer")
        if micro_step < Data2DBConst.MIN_MICRO_STEP or micro_step > Data2DBConst.MAX_MICRO_STEP:
            raise ValueError(f"Micro step must be between"
                             f"{Data2DBConst.MIN_MICRO_STEP} and {Data2DBConst.MAX_MICRO_STEP}")


def load_mapping(mapping_path):
    if mapping_path and isinstance(mapping_path, str):
        return load_json(mapping_path)
    else:
        return {}


@dataclass
class TensorProcessingParams:
    """Tensor处理函数的参数载体"""
    tensor_data: Dict[str, Any]
    target_prefix: str
    vpp_stage: int
    micro_step: int
    step: int
    rank: int
    metric_id: int
    metric_type: Optional[str]


class DumpRecordBuilder:
    def __init__(self, db: DumpDB, data_dir, mapping, micro_step):
        self.db = db
        self.data_dir = data_dir
        self.mapping = mapping
        self.micro_step = micro_step if micro_step else None

    @staticmethod
    def extract_target_info(full_key):
        """
        从完整的key中提取target信息
        格式: Module.layerN.operation.M
        """
        vpp_stage = 0  # 暂不提取vpp信息
        micro_step = 0
        target_prefix = full_key

        parts = full_key.split(Const.SEP)
        if parts and parts[-1].isdigit():
            micro_step = int(parts[-1])
            # 重新构建target_prefix（去掉最后一个数字部分）
            target_prefix = Const.SEP.join(parts[:-1])

        return target_prefix, vpp_stage, micro_step

    @staticmethod
    def parse_tensor_target(metric_type, tensor_type, tensor_idx):
        """根据tensor类型和索引生成target后缀"""
        if metric_type in [Data2DBConst.FORWARD, Data2DBConst.RECOMPUTE]:
            if tensor_type == Const.INPUT_ARGS:
                return f".input.{tensor_idx}"
            elif tensor_type == Const.OUTPUT:
                return f".output.{tensor_idx}"
            elif tensor_type == Const.PARAMS:
                return f".parameters.{tensor_idx}"
        elif metric_type == Data2DBConst.BACKWARD:
            if tensor_type == Const.INPUT:
                return f".input.{tensor_idx}"
            elif tensor_type == Const.OUTPUT:
                return f".output.{tensor_idx}"
        return ""

    @staticmethod
    def process_tensor_value(value):
        """处理统计量值  将inf转换为float极值"""
        if value is None:
            return sys.float_info.max
        elif value == float("inf"):
            return sys.float_info.max - 1  # 最大float值
        elif value == float("-inf"):
            return sys.float_info.min + 1  # 最小float值
        elif isinstance(value, (int, float)):
            return float(value)
        else:
            return None

    def import_data(self):
        """导入数据"""
        logger.info("Scanning data directory...")

        # 扫描所有step和rank
        max_rank = 0
        min_step = float('inf')
        max_step = 0
        valid_ranks = defaultdict(list)

        # 首先扫描目录结构获取step和rank信息
        for step_dir in os.listdir(self.data_dir):
            if not step_dir.startswith('step'):
                continue
            try:
                step = int(step_dir[4:])  # 提取step数字
            except ValueError:
                continue
            min_step = min(min_step, step *
                           self.micro_step if self.micro_step else step)
            max_step = max(max_step, (step + 1) *
                           self.micro_step if self.micro_step else step)

            step_path = os.path.join(self.data_dir, step_dir)
            for rank_dir in os.listdir(step_path):
                if not rank_dir.startswith('rank'):
                    continue
                if rank_dir == "rank":
                    rank = 0
                else:
                    try:
                        rank = int(rank_dir[4:])  # 提取rank数字
                    except ValueError:
                        continue

                rank_path = os.path.join(step_path, rank_dir)
                json_path = os.path.join(rank_path, "dump.json")
                if os.path.exists(json_path):
                    valid_ranks[step].append((rank, json_path))
                    max_rank = max(max_rank, rank)

        if min_step == float('inf'):
            logger.warning(f"No valid step directories found in: {self.data_dir},"
                           f"looking for directories starting with 'step' (e.g., step0, step1")
            return

        # 更新全局统计信息
        global_stats = {
            "max_rank": max_rank,
            "min_step": min_step,
            "max_step": max_step
        }
        for metric_name in Data2DBConst.METRICS:
            global_stats[metric_name] = Data2DBConst.ORDERED_STAT
        self.db.init_global_stats_data(global_stats)

        for step in tqdm(sorted(valid_ranks.keys()), desc="Processing steps"):
            step_dir = f"step{step}"
            step_path = os.path.join(self.data_dir, step_dir)

            check_file_or_directory_path(step_path, isdir=True)

            # 为当前step的所有有效rank创建进度条
            if not valid_ranks[step]:
                continue
            for rank, json_path in tqdm(valid_ranks[step], desc=f"Step {step} ranks", leave=False):
                self._process_dump_file(
                    json_path, step, rank)
        self.db.extract_tags_from_processed_targets()

    def _determine_metric_type(self, full_key, tensor_data):
        for key, value in self.mapping.items():
            full_key = full_key.replace(key, value)

        # 确定metric类型
        metric_type = None
        if f"{Const.SEP}{Const.FORWARD}{Const.SEP}" in full_key or \
                full_key.endswith(f"{Const.SEP}{Const.FORWARD}"):
            if tensor_data.get('is_recompute', False):
                metric_type = Data2DBConst.RECOMPUTE
            else:
                metric_type = Data2DBConst.FORWARD
            full_key = full_key.replace(
                f"{Const.SEP}{Const.FORWARD}{Const.SEP}", Const.SEP)  # module
            full_key = full_key.replace(
                f"{Const.SEP}{Const.FORWARD}", "")  # api
        elif f"{Const.SEP}{Const.BACKWARD}{Const.SEP}" in full_key or \
                full_key.endswith(f"{Const.SEP}{Const.BACKWARD}"):
            metric_type = Data2DBConst.BACKWARD
            full_key = full_key.replace(
                f"{Const.SEP}{Const.BACKWARD}{Const.SEP}", Const.SEP)
            full_key = full_key.replace(f"{Const.SEP}{Const.BACKWARD}", "")
        elif full_key.endswith(Const.PARAMS_GRAD):
            metric_type = Data2DBConst.PARAMETERS_GRAD
            full_key = full_key.replace(f"{Const.SEP}{Const.PARAMS_GRAD}", "")
        # fsdp
        elif len(full_key.split(Const.SEP)) >= 3:
            parts = full_key.split(Const.SEP)
            if parts[-2] == Const.PARAMS_GRAD:
                metric_type = Data2DBConst.PARAMETERS_GRAD
                full_key = Const.SEP.join(parts[:-2])

        return metric_type, full_key

    def _add_tensor_data(self, tensor, target_name, tensor_params: TensorProcessingParams, batch_data):
        """添加tensor数据方法"""

        cache_key = (target_name, tensor_params.vpp_stage,
                     tensor_params.micro_step)

        # 如果缓存中不存在，创建临时ID
        cache_id_dict = self.db.cache_targets(
            cache_key, tensor_params.metric_id)

        # 准备数据行, 这里id还是个临时id, 需要更新后读取, 第三个实际为{"id": 0}
        row_data = [tensor_params.rank, tensor_params.step, cache_id_dict, tensor_params.metric_id]
        for stat in Data2DBConst.ORDERED_STAT:
            value = tensor.get(stat.capitalize(), None)
            row_data.append(DumpRecordBuilder.process_tensor_value(value))
        batch_data.append((row_data))

    def _process_forward_data(self, tensor_params: TensorProcessingParams, batch_data):
        """处理forward/recompute数据"""
        # 处理input_args
        if Const.INPUT_ARGS in tensor_params.tensor_data:
            for idx, tensor in enumerate(tensor_params.tensor_data[Const.INPUT_ARGS]):
                if isinstance(tensor, dict) and tensor.get(Const.TYPE) in Data2DBConst.SUPPORT_TYPE:
                    if tensor.get(Const.DTYPE) not in Data2DBConst.SUPPORT_DTYPE:
                        continue
                    target_suffix = DumpRecordBuilder.parse_tensor_target(
                        Data2DBConst.FORWARD, Const.INPUT_ARGS, idx)
                    target_name = tensor_params.target_prefix + target_suffix
                    self._add_tensor_data(
                        tensor, target_name, tensor_params, batch_data)

        # 处理output
        if Const.OUTPUT in tensor_params.tensor_data:
            for idx, tensor in enumerate(tensor_params.tensor_data[Const.OUTPUT]):
                if isinstance(tensor, dict) and tensor.get(Const.TYPE) in Data2DBConst.SUPPORT_TYPE:
                    if tensor.get(Const.DTYPE) not in Data2DBConst.SUPPORT_DTYPE:
                        continue
                    target_suffix = DumpRecordBuilder.parse_tensor_target(
                        Data2DBConst.FORWARD, Const.OUTPUT, idx)
                    target_name = tensor_params.target_prefix + target_suffix
                    self._add_tensor_data(
                        tensor, target_name, tensor_params, batch_data)

        # 处理parameters
        if Const.PARAMS in tensor_params.tensor_data:
            for param_name, param_tensors in tensor_params.tensor_data[Const.PARAMS].items():
                if not (isinstance(param_tensors, list) and param_tensors and isinstance(param_tensors[0], dict)):
                    continue
                if param_tensors[0].get(Const.TYPE) in Data2DBConst.SUPPORT_TYPE:
                    if param_tensors[0].get(Const.DTYPE) not in Data2DBConst.SUPPORT_DTYPE:
                        continue
                    target_suffix = DumpRecordBuilder.parse_tensor_target(
                        Data2DBConst.FORWARD, Const.PARAMS, param_name)
                    target_name = tensor_params.target_prefix + target_suffix
                    self._add_tensor_data(
                        param_tensors[0], target_name, tensor_params, batch_data)

    def _process_parameters_data(self, tensor_params: TensorProcessingParams, batch_data):
        """处理parameters数据"""
        for param_name, param_tensors in tensor_params.tensor_data.items():
            if not (isinstance(param_tensors, list) and param_tensors and isinstance(param_tensors[0], dict)):
                continue
            if param_tensors[0].get(Const.TYPE) in Data2DBConst.SUPPORT_TYPE:
                if param_tensors[0].get(Const.DTYPE) not in Data2DBConst.SUPPORT_DTYPE:
                    continue
                target_name = tensor_params.target_prefix + f".{param_name}"
                self._add_tensor_data(
                    param_tensors[0], target_name, tensor_params, batch_data)

    def _process_backward_data(self, tensor_params: TensorProcessingParams, batch_data):
        """处理backward数据"""
        # 处理input
        if Const.INPUT in tensor_params.tensor_data:
            for idx, tensor in enumerate(tensor_params.tensor_data[Const.INPUT]):
                if isinstance(tensor, dict) and tensor.get(Const.TYPE) in Data2DBConst.SUPPORT_TYPE:
                    if tensor.get(Const.DTYPE) not in Data2DBConst.SUPPORT_DTYPE:
                        continue
                    target_suffix = DumpRecordBuilder.parse_tensor_target(
                        Data2DBConst.BACKWARD, Const.INPUT, idx)
                    target_name = tensor_params.target_prefix + target_suffix
                    self._add_tensor_data(
                        tensor, target_name, tensor_params, batch_data)

        # 处理output
        if Const.OUTPUT in tensor_params.tensor_data:
            for idx, tensor in enumerate(tensor_params.tensor_data[Const.OUTPUT]):
                if isinstance(tensor, dict) and tensor.get(Const.TYPE) in Data2DBConst.SUPPORT_TYPE:
                    if tensor.get(Const.DTYPE) not in Data2DBConst.SUPPORT_DTYPE:
                        continue
                    target_suffix = DumpRecordBuilder.parse_tensor_target(
                        Data2DBConst.BACKWARD, Const.OUTPUT, idx)
                    target_name = tensor_params.target_prefix + target_suffix
                    self._add_tensor_data(
                        tensor, target_name, tensor_params, batch_data)

    def _process_dump_file(self, json_path, step, rank):
        """处理单个dump.json文件"""
        data = load_json(json_path)
        if 'data' not in data or not isinstance(data['data'], dict):
            return

        batch_data = []
        # 预先计算所有metric_type的table_name

        for i, (ori_key, tensor_data) in enumerate(data['data'].items()):
            if not isinstance(ori_key, str) or not isinstance(tensor_data, dict):
                continue
            metric_type, full_key = self._determine_metric_type(
                ori_key, tensor_data)
            if not metric_type:
                continue

            # 使用缓存获取metric_id和stats
            metric_id = self.db.get_metric_id(metric_type)
            target_prefix, vpp_stage, mstep = DumpRecordBuilder.extract_target_info(
                full_key)
            if self.micro_step:
                current_mstep = step * self.micro_step + mstep
                mstep = 0
            else:
                current_mstep = step

            tensor_params = TensorProcessingParams(
                tensor_data=tensor_data,
                target_prefix=target_prefix,
                vpp_stage=vpp_stage,
                micro_step=mstep,
                step=current_mstep,
                rank=rank,
                metric_id=metric_id,
                metric_type=metric_type
            )

            # 处理不同类型的tensor数据
            if metric_type in [Data2DBConst.FORWARD, Data2DBConst.RECOMPUTE]:
                self._process_forward_data(tensor_params, batch_data)
            elif metric_type == Data2DBConst.PARAMETERS_GRAD:
                self._process_parameters_data(tensor_params, batch_data)
            elif metric_type == Data2DBConst.BACKWARD:
                self._process_backward_data(tensor_params, batch_data)

            if len(batch_data) % Data2DBConst.BATCH_SIZE == 0 and len(batch_data) > 0:
                self.db.batch_insert_targets()
                self.db.batch_insert_data(batch_data)
                batch_data = []

        self.db.batch_insert_targets()
        self.db.batch_insert_data(batch_data)


def _data2db_service_parser(parser):
    parser.add_argument('--db', type=str, required=True,
                        help='Path to SQLite database file')
    parser.add_argument('--data', type=str, required=True,
                        help='Path to dump output directory')
    parser.add_argument('--mapping', type=str, default=None,
                        help='Path to optional JSON mapping file')
    parser.add_argument('--micro_step', type=int, default=None,
                        help='Specific micro step value to split data in one step (must be between 1 and 10000)')


def _data2db_command(args):
    data_path = args.data

    check_file_or_directory_path(data_path, isdir=True, is_strict=True)
    create_directory(args.db)
    db_path = os.path.join(args.db, "monitor_metrics.db")

    if os.path.exists(db_path):
        logger.warning(f"Existing path {db_path} will be recovered")
        remove_path(db_path)

    micro_step = args.micro_step
    validate_micro_step(micro_step)
    mapping = load_mapping(args.mapping)
    db = DumpDB(db_path)
    builder = DumpRecordBuilder(
        db, data_path, mapping=mapping, micro_step=micro_step)
    builder.import_data()

    recursive_chmod(args.db)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import numpy as np

from msprobe.core.common.log import logger
from msprobe.core.compare.npy_compare import CompareOps



def in_different_shape(a, b):
    if a.shape != b.shape:
        logger.warning(f"a, b are in different shape. a: {a.shape}, b: {b.shape}")
        return True
    return False


def l2_distance(a, b):
    if a is None or b is None:
        return None
    if in_different_shape(a, b):
        return None
    return np.linalg.norm(a - b).item()


def cos_sim(a, b):
    if a is None or b is None:
        return None

    if in_different_shape(a, b):
        return None
    if a.ndim > 0:
        a = a.flatten().squeeze()
        b = b.flatten().squeeze()

    num = a.dot(b)
    a_norm = np.linalg.norm(a)
    b_norm = np.linalg.norm(b)

    if a_norm == 0 and b_norm == 0:
        return 1.
    if a_norm == 0 or b_norm == 0:
        logger.warning(f'One tensor norm is zero.')
        return None

    sim = num / (a_norm * b_norm)

    return sim.item()


def numel(a, b):
    n1 = a.size
    n2 = b.size
    if n1 != n2:
        logger.warning('parameters have different number of element')
        return (n1, n2)
    return n1


def shape(a, b):
    if in_different_shape(a, b):
        return [list(a.shape), list(b.shape)]
    return list(a.shape)


METRIC_FUNC = {
    'l2': l2_distance, 
    'cos': cos_sim, 
    'numel': numel, 
    'shape': shape
    }



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import threading
from collections import defaultdict

import mindspore as ms
from mindspore import nn

from msprobe.mindspore.common.utils import is_mindtorch, register_backward_hook_functions

ms_version = ms.__version__


def add_cell_count(name):
    HOOKCell.cell_count[name] += 1


def get_cell_count(name):
    return HOOKCell.cell_count[name]


def __init__(self, hook_build_func) -> None:
    super(HOOKCell, self).__init__()
    self.msprobe_input_kwargs = {}
    prefix = self.prefix_api_name if hasattr(self, "prefix_api_name") else ""
    if callable(hook_build_func):
        hook_set = hook_build_func(prefix)
        if ms_version < "2.6.0" and not is_mindtorch():
            getattr(self, "_forward_pre_hook", {})[id(self)] = hook_set.forward_pre_hook
            if hook_set.forward_hook:
                getattr(self, "_forward_hook", {})[id(self)] = hook_set.forward_hook
        else:
            self.register_forward_pre_hook(hook_set.forward_pre_hook)
            if hook_set.forward_hook:
                self.register_forward_hook(hook_set.forward_hook)


def __call__(self, *args, **kwargs):
    tid = threading.get_ident()
    self.msprobe_input_kwargs[tid] = kwargs
    out = super(HOOKCell, self).__call__(*args, **kwargs)
    return out


hook_cell_dict = {
    "cell_count": defaultdict(int),
    "add_cell_count": staticmethod(add_cell_count),
    "get_cell_count": staticmethod(get_cell_count),
    "__init__": __init__,
    "__call__": __call__
}

if is_mindtorch():
    import torch
    HOOKCell = type("HOOKCell", (torch.nn.Module,), hook_cell_dict)
else:
    HOOKCell = type("HOOKCell", (nn.Cell,), hook_cell_dict)



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
NpuVsNpuComparison class. This class mainly involves the compare function.
"""
import numpy as np

from msprobe.msaccucmp.cmp_utils import utils
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.dump_parse.dump import CompareData
from msprobe.msaccucmp.algorithm_manager.algorithm_manager import AlgorithmManager
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import FusionOp
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import Tensor
from msprobe.msaccucmp.vector_cmp.fusion_manager import compare_result
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.overflow.overflow_detection import OverflowDetection
from msprobe.msaccucmp.dump_parse.ffts_parser import FFTSParser
from msprobe.msaccucmp.dump_parse import dump_utils
from msprobe.msaccucmp.conversion.tensor_conversion import ConvertSingleTensorFormat 


class NpuVsNpuComparison:
    """
    The class for npu vs npu comparison
    """

    def __init__(self: any, compare_data: CompareData, fusion_op_list: list, algorithm_manager: AlgorithmManager,
                 overflow_detection: bool = False) -> None:
        self.compare_data = compare_data
        self.fusion_op_list = fusion_op_list
        self.algorithm_manager = algorithm_manager
        self.op_name = fusion_op_list[0].op_name
        self.overflow_detection = overflow_detection
        self.enable_padding_restore = True # 预留用于控制是否开启补齐恢复
        self._tensor_converter = ConvertSingleTensorFormat()

    def check_tensor_valid(self: any, my_output_tensor_list: any, ground_truth_tensor_list: any,
                           tensor_type: str) -> (int, str):
        """
        check tensor valid
        """
        # check the length is same
        if len(my_output_tensor_list) != len(ground_truth_tensor_list):
            message = log.print_not_match_error(
                self.op_name, 'number of %s' % tensor_type, str(len(my_output_tensor_list)),
                str(len(ground_truth_tensor_list)))
            return CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message
        if len(my_output_tensor_list) != 0:
            # check each tensor format and shape valid
            return self._check_op_data_valid(my_output_tensor_list, ground_truth_tensor_list, tensor_type)
        message = '[%s] There is no %s. Skip the %s:%s.' % (self.op_name, tensor_type, self.op_name, tensor_type)
        log.print_info_log(message)
        return CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message

    def _restore_tensor_data_if_needed(self, tensor: Tensor):
        """
        根据开关决定是否对 tensor 做格式转换 + 按 original_shape 切掉 padding。
        """
        if not self._tensor_converter:
            return tensor.data

        try:
            # ConvertSingleTensorFormat.__call__ 返回的是 np.ndarray
            restored = self._tensor_converter(tensor)
        except CompareError as ee:
            # 出错时回退到原始数据，保证比对流程不中断
            log.print_error_log(ee)
            return tensor.data
        except Exception as ee:  # 兜底
            log.print_error_log(ee)
            return tensor.data

        return restored

    def compare(self: any) -> (int, bool, list):
        """
        Compare for npu vs npu by op_name
        :return ret: return code
        :return dump_match: True, at least one operator match;False, no operator match
        :return result: the compare result by the fusion op list
        """
        single_op_cmp_result = compare_result.SingleOpCmpResult()
        if len(self.fusion_op_list) == 1:
            return self._make_one_dump_file_result()
        # get my output and ground truth tensor
        error_msg = []
        try:
            my_output_dump_data = self._get_dump_data(
                self.fusion_op_list[0], self.compare_data.left_dump_info.path,
                self.compare_data.left_dump_info.op_name_to_task_mode_map, ConstManager.LEFT_TYPE)
        except CompareError as error:
            error_msg.append(error.message)
            fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager)
            _result = fusion_op_result.get_result(self.fusion_op_list[0], None, error_msg)
            result_info = utils.ResultInfo(
                self.fusion_op_list[0].op_name, True, _result.result_list, error.code,
                [], _result.input_result_list, _result.output_result_list, _result.is_ffts,
                {}, True)
            single_op_cmp_result.update_attr(result_info)
            return error.code, True, [single_op_cmp_result]

        ground_truth_dump_data = self._get_dump_data(
            self.fusion_op_list[1], self.compare_data.right_dump_info.path,
            self.compare_data.right_dump_info.op_name_to_task_mode_map, ConstManager.RIGHT_TYPE)

        compare_vector_result = []
        # check npu input data valid
        input_ret, input_error_msg = self.check_tensor_valid(
            my_output_dump_data.data.input_data, ground_truth_dump_data.data.input_data, ConstManager.INPUT)
        if input_ret == CompareError.MSACCUCMP_NONE_ERROR:
            # compare input
            compare_vector_result += self._compare_by_tensor(my_output_dump_data, ground_truth_dump_data,
                                                             ConstManager.INPUT)

        # check npu output data valid
        output_ret, output_error_msg = self.check_tensor_valid(
            my_output_dump_data.data.output_data, ground_truth_dump_data.data.output_data, ConstManager.OUTPUT)

        if output_ret == CompareError.MSACCUCMP_NONE_ERROR:
            # compare output
            compare_vector_result += self._compare_by_tensor(my_output_dump_data, ground_truth_dump_data,
                                                             ConstManager.OUTPUT)

        if not my_output_dump_data.data.ffts_file_check:
            msg = "This is a FFTS+ mode dump data, The number of files does not match the number of thread"
            error_msg.append(msg)
        # if no input and output, result is NaN
        if input_ret != CompareError.MSACCUCMP_NONE_ERROR and output_ret != CompareError.MSACCUCMP_NONE_ERROR:
            error_msg.append(input_error_msg)
            error_msg.append(output_error_msg)
            compare_vector_result = None
        else:
            output_ret = CompareError.MSACCUCMP_NONE_ERROR
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager)
        _result = fusion_op_result.get_result(self.fusion_op_list[0], compare_vector_result, error_msg)

        result_info = utils.ResultInfo(
            my_output_dump_data.name, True, _result.result_list, output_ret,
            [], _result.input_result_list, _result.output_result_list, _result.is_ffts,
            {}, True)

        single_op_cmp_result.update_attr(result_info)

        return output_ret, True, [single_op_cmp_result]

    def _make_one_dump_file_result(self: any) -> (int, bool, list):
        error_msg = []
        # if only left or right has dump file, the result is NaN
        single_op_cmp_result = compare_result.SingleOpCmpResult()
        if self.fusion_op_list[0].op_type == ConstManager.LEFT_TYPE:
            message = '[%s] There is no the ground truth dump file for the op "%s".' % (self.op_name, self.op_name)
            log.print_warn_log(message)
            error_msg.append(message)
        elif self.fusion_op_list[0].op_type == ConstManager.RIGHT_TYPE:
            message = '[%s] There is no the my output dump file for the op "%s".' % (self.op_name, self.op_name)
            log.print_warn_log(message)
            error_msg.append(message)
        fusion_op_result = compare_result.FusionOpComResult(self.algorithm_manager,
                                                            overflow_detection=self.overflow_detection)
        _result = fusion_op_result.get_result(self.fusion_op_list[0], None, error_msg, no_dump_file=True)

        result_info = utils.ResultInfo(
            self.fusion_op_list[0].op_name, False, _result.result_list,
            CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR,
            self.fusion_op_list[0].input_list, _result.input_result_list,
            _result.output_result_list, _result.is_ffts, {}, True)

        single_op_cmp_result.update_attr(result_info)

        return CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR, False, [single_op_cmp_result]

    def _get_dump_data(self: any, fusion_op: FusionOp, dump_path: str,
                       op_name_to_task_mode_map, dump_type: str) -> Tensor:
        """
        get dump data by fusion op output_desc
        """
        dump_file_list = fusion_op.output_desc
        if not dump_file_list:
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR)
        dump_data_list = [dump_utils.parse_dump_file(dump_file_path, self.compare_data.dump_version)
                          for dump_file_path in dump_file_list]
        dump_mode = op_name_to_task_mode_map.get(self.op_name)
        if dump_mode == ConstManager.AUTOMATIC_MODE or dump_mode == ConstManager.MANUAL_MODE:
            ffts_parser = FFTSParser(dump_file_list, dump_data_list)
            dump_file_path, dump_data = ffts_parser.parse_ffts
            log.print_info_log(
                'The "%s" in the path "%s" is FFTS+ dump data. After process the output data, the file path is "%s".'
                % (fusion_op.op_name, dump_path, dump_file_path))
        else:
            dump_file_path = dump_file_list[-1]
            dump_data = dump_data_list[-1]
        log.print_info_log('[%s] [%s] %s' % (fusion_op.op_name, dump_type, dump_file_path))
        if dump_data.op_name and dump_data.attr:
            fusion_op.op_name = dump_data.op_name
        tensor = Tensor(fusion_op.op_name, 0, '', [])
        tensor.set_path(dump_file_path)
        tensor.set_data(dump_data)
        return tensor

    def _check_op_data_valid(self: any, my_output_list: any, ground_truth_list: any, tensor_type: str) -> (int, str):
        """
        check format and shape of each tensor valid
        """
        message = ""
        tensor_id_prefix = "%s:%s" % (self.op_name, tensor_type)
        for index, (my_output_tensor, ground_truth_tensor) in enumerate(zip(my_output_list, ground_truth_list)):
            tensor_id = '%s:%d' % (tensor_id_prefix, index)
            # check format valid
            if my_output_tensor.tensor_format != ground_truth_tensor.tensor_format:
                message = log.print_not_match_error(
                    self.op_name, 'format',
                    common.get_format_string(my_output_tensor.tensor_format),
                    common.get_format_string(ground_truth_tensor.tensor_format), tensor_id)
                return CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message

            # check the length of shape is the same
            if len(my_output_tensor.shape) != len(ground_truth_tensor.shape):
                message = log.print_not_match_error(
                    self.op_name, 'shape',
                    utils.convert_shape_to_string(my_output_tensor.shape),
                    utils.convert_shape_to_string(ground_truth_tensor.shape), tensor_id)
                return CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message
            # check each dim in shape is the same
            for my_output_dim, ground_truth_dim in zip(my_output_tensor.shape, ground_truth_tensor.shape):
                if my_output_dim != ground_truth_dim:
                    message = log.print_not_match_error(
                        self.op_name, 'shape',
                        utils.convert_shape_to_string(my_output_tensor.shape),
                        utils.convert_shape_to_string(ground_truth_tensor.shape), tensor_id)
                    return CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message
        return CompareError.MSACCUCMP_NONE_ERROR, message

    def _compare_by_one_tensor(self: any, my_output_dump_data: Tensor, ground_truth_dump_data: Tensor,
                               my_output_tensor: any, ground_truth_tensor: any) -> (list, list):
        error_msg = []
        tensor_id = f"{self.op_name}_TENSOR"

        # 1. deserialize output data to array
        if my_output_tensor and ground_truth_tensor:
            if self.enable_padding_restore:

                restored_left = self._restore_tensor_data_if_needed(my_output_tensor)
                restored_right = self._restore_tensor_data_if_needed(ground_truth_tensor)

                # 强制 numpy flatten，避免后端 compare 出现 numpy bool 错误
                my_output_data_array = np.asarray(restored_left).astype(np.float32).flatten()
                ground_truth_data_array = np.asarray(restored_right).astype(np.float32).flatten()

                # 若长度不一致，直接报 warning，方便定位问题
                if my_output_data_array.shape != ground_truth_data_array.shape:
                    message = f"[{tensor_id}] Shape mismatch after restore: " \
                              f"{my_output_data_array.shape} vs {ground_truth_data_array.shape}"
                    log.print_warn_log(message)
                    raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR, message)
            else:
                my_output_data_array = my_output_tensor.data.flatten()
                ground_truth_data_array = ground_truth_tensor.data.flatten()
        else:
            return self.algorithm_manager.make_nan_result(), error_msg
        
        try:
            # 2. compare by support algorithm
            algorithm_result, error_msg = self.algorithm_manager.compare(
                my_output_data_array, ground_truth_data_array,
                {'my_output_dump_file': my_output_dump_data.path,
                 'ground_truth_dump_file': ground_truth_dump_data.path,
                 'shape_type': utils.get_shape_type(my_output_tensor.shape)})
        except CompareError as compare_error:
            if isinstance(compare_error, CompareError):
                error_msg.append(compare_error.message)
            algorithm_result = self.algorithm_manager.make_nan_result()

        return algorithm_result, error_msg

    def _compare_by_tensor(self: any, my_output_dump_data: Tensor, ground_truth_dump_data: Tensor,
                           tensor_type: str) -> list:
        tensor_result_list = []
        if tensor_type == ConstManager.INPUT:
            my_output_tensor_list = my_output_dump_data.data.input_data
            ground_truth_tensor_list = ground_truth_dump_data.data.input_data
            is_input = True
        else:
            my_output_tensor_list = my_output_dump_data.data.output_data
            ground_truth_tensor_list = ground_truth_dump_data.data.output_data
            is_input = False
        # compare each tensor
        for index, (my_output_tensor, ground_truth_tensor) in enumerate(
                zip(my_output_tensor_list, ground_truth_tensor_list)):
            tensor_id = '%s:%s:%d' % (my_output_dump_data.name, tensor_type, index)
            log.print_info_log('[%s] compare %s %s for %s.'
                               % (self.fusion_op_list[0].op_name,
                                  common.get_format_string(my_output_tensor.tensor_format),
                                  utils.convert_shape_to_string(my_output_tensor.shape),
                                  tensor_id))
            algorithm_result, error_msg = self._compare_by_one_tensor(my_output_dump_data, ground_truth_dump_data,
                                                                      my_output_tensor, ground_truth_tensor)
            # Check whether the current input/output data overflows
            overflow_result = ''
            if self.overflow_detection:
                overflow_result = OverflowDetection.process_model_overflow_detection(my_output_dump_data.name,
                                                                                     index, is_input, my_output_tensor)
            my_output_tensor_dtype = utils.get_data_type(my_output_tensor.data_type)
            ground_truth_tensor_dtype = utils.get_data_type(ground_truth_tensor.data_type)
            my_output_tensor_address = utils.get_address_from_tensor(my_output_tensor)
            ground_truth_tensor_address = utils.get_address_from_tensor(ground_truth_tensor)
            op_type = dump_utils.get_op_type_from_file_name(my_output_dump_data.path)

            # 3. merge result
            tensor_info = {
                "tensor_id": tensor_id,
                "shape": my_output_tensor.shape,
                "op_type": op_type,
                "my_output_dtype": my_output_tensor_dtype,
                "ground_truth_dtype": ground_truth_tensor_dtype,
                "my_output_address": my_output_tensor_address,
                "ground_truth_address": ground_truth_tensor_address
            }
            tensor_result_list.append(
                compare_result.TensorResult(
                    tensor_info, [algorithm_result, overflow_result], error_msg, my_output_tensor.is_ffts))
        return tensor_result_list



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
import sys
import json
import os
from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import check_file_or_directory_path


def _detect_framework_from_api_info(api_info_path: str) -> str:
    """从 -api_info 指定的 dump.json 中读取 framework 字段."""
    check_file_or_directory_path(api_info_path, False)
    if not api_info_path:
        raise ValueError("Argument -api_info is required to detect framework.")

    if not os.path.exists(api_info_path):
        raise FileNotFoundError(f"api_info file does not exist: {api_info_path}")

    with open(api_info_path, "r", encoding="utf-8") as f:
        try:
            info = json.load(f)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse JSON from {api_info_path}: {e}") from e

    framework = info.get("framework")
    if not framework:
        raise ValueError(f'Key "framework" not found in {api_info_path}')

    framework = str(framework).lower()
    # 统一映射到 Const 中定义的名字，方便复用其他逻辑
    if framework in ("pytorch", "pt", Const.PT_FRAMEWORK.lower()):
        return Const.PT_FRAMEWORK
    if framework in ("mindspore", "ms", Const.MS_FRAMEWORK.lower(), Const.MT_FRAMEWORK.lower()):
        return Const.MS_FRAMEWORK

    raise ValueError(f"Unsupported framework in api_info: {framework}")


def acc_check_cli(argv):
    """
    msprobe acc_check ... 的统一入口。
    1. 先解析出 -api_info
    2. 根据 dump.json 中的 framework 动态选择 PT/MS 的 parser + 命令。
    """
    # 第一阶段：只解析 -api_info，其他参数先不管
    pre_parser = argparse.ArgumentParser(add_help=False)
    pre_parser.add_argument("-api_info", required=True, help="Path to api_info json file.")
    pre_args, _ = pre_parser.parse_known_args(argv)

    framework = _detect_framework_from_api_info(pre_args.api_info)

    if framework == Const.PT_FRAMEWORK:
        # PyTorch 路径：使用原来的 PT acc_check 实现
        from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check import _acc_check_parser, acc_check_command

        pt_parser = argparse.ArgumentParser(
            prog="msprobe acc_check",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="Run PyTorch acc_check with msprobe."
        )
        _acc_check_parser(pt_parser)  # 这里会给 parser 加上原来的所有 PT acc_check 参数（包括 -api_info）

        pt_args = pt_parser.parse_args(argv)
        acc_check_command(pt_args)

    elif framework == Const.MS_FRAMEWORK:
        # MindSpore 路径：复用原来的 MS api_checker_main
        from msprobe.mindspore.api_accuracy_checker.cmd_parser import add_api_accuracy_checker_argument
        from msprobe.mindspore.api_accuracy_checker.main import api_checker_main

        ms_parser = argparse.ArgumentParser(
            prog="msprobe acc_check",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="Run MindSpore Check  with msprobe."
        )
        add_api_accuracy_checker_argument(ms_parser)  # 给 acc_check 的 parser 加上原来 MS 的所有参数

        ms_args = ms_parser.parse_args(argv)
        api_checker_main(ms_args)


def multi_acc_check_cli(argv):
    """
    msprobe multi_acc_check ... 的统一入口。
    同样通过 -api_info -> dump.json -> framework 做分发。
    """
    pre_parser = argparse.ArgumentParser(add_help=False)
    pre_parser.add_argument("-api_info", required=True, help="Path to api_info json file.")
    pre_args, _ = pre_parser.parse_known_args(argv)

    framework = _detect_framework_from_api_info(pre_args.api_info)

    if framework == Const.PT_FRAMEWORK:
        # PyTorch 多进程路径：沿用原来的 prepare_config + run_parallel_ut
        from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check import _acc_check_parser
        from msprobe.pytorch.api_accuracy_checker.acc_check.multi_acc_check import prepare_config, run_parallel_ut

        pt_parser = argparse.ArgumentParser(
            prog="msprobe multi_acc_check",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="Run PyTorch acc_check in parallel with msprobe."
        )
        _acc_check_parser(pt_parser)
        pt_parser.add_argument(
            "-n", "--num_splits",
            type=int,
            choices=range(1, 65),
            default=8,
            help="Number of splits for parallel processing. Range: 1-64"
        )

        pt_args = pt_parser.parse_args(argv)
        config = prepare_config(pt_args)
        run_parallel_ut(config)

    elif framework == Const.MS_FRAMEWORK:
        # MindSpore 多进程路径：沿用原来的 multi_add_api_accuracy_checker_argument + mul_api_checker_main
        from msprobe.mindspore.api_accuracy_checker.cmd_parser import multi_add_api_accuracy_checker_argument
        from msprobe.mindspore.api_accuracy_checker.main import mul_api_checker_main

        ms_parser = argparse.ArgumentParser(
            prog="msprobe multi_acc_check",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="Run MindSpore Check in parallel with msprobe."
        )
        multi_add_api_accuracy_checker_argument(ms_parser)

        ms_args = ms_parser.parse_args(argv)
        mul_api_checker_main(ms_args)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import namedtuple
import numpy as np
import torch

from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_config import StandardConfig
from msprobe.pytorch.api_accuracy_checker.precision_standard.base_standard import BaseCompare, BasePrecisionCompare
from msprobe.core.common.const import Const, CompareConst
from msprobe.pytorch.api_accuracy_checker.compare.algorithm import calc_ratio, get_ulp_err
from msprobe.pytorch.api_accuracy_checker.compare.compare_utils import ApiPrecisionCompareColumn, check_inf_or_nan, \
    is_inf_or_nan
from msprobe.pytorch.api_accuracy_checker.common.utils import is_dtype_fp8_or_hif8


UlpInfNanConsistency = namedtuple('UlpInfNanConsistency', ['mean_ulp_err_inf_nan_consistency', 
                                             'ulp_err_proportion_ratio_inf_nan_consistency'])


class UlpCompare(BaseCompare):
    """
    Ulp compare comparison class for calculating accuracy metrics.

    Attributes:
        bench_output (array-like): The benchmark output values.
        device_output (array-like): The device output values.
        dtype (torch.dtype): The data type of the outputs (e.g., torch.float32 or torch.float16).
        ulp_err (array-like): The ULP errors calculated from the benchmark and device outputs.

    Methods:
        _stat_max_ulp_err(ulp_err): Calculates the maximum ULP error.
        _stat_mean_ulp_err(ulp_err): Calculates the mean ULP error.
        _stat_ulp_error_proportion(ulp_err): Calculates the proportion of ULP errors exceeding a threshold.
        _pre_compare(): Prepares for comparison by calculating ULP errors.
        _compute_metrics(): Computes the ULP error metrics.
    """
    def __init__(self, input_data):
        super(UlpCompare, self).__init__(input_data)
    
    @staticmethod
    def _stat_max_ulp_err(ulp_err):
        return np.max(ulp_err)
    
    @staticmethod
    def _stat_mean_ulp_err(ulp_err):
        return np.mean(ulp_err)
    
    def _stat_ulp_error_proportion(self, ulp_err):
        if self.dtype == torch.float32:
            return np.sum(ulp_err > CompareConst.ULP_FLOAT32_THRESHOLD) / self.bench_output.size
        else:
            return np.sum(ulp_err > CompareConst.ULP_FLOAT16_THRESHOLD) / self.bench_output.size
    
    def _pre_compare(self):
        self.ulp_err = get_ulp_err(self.bench_output, self.device_output, self.dtype)
    
    def _compute_metrics(self):
        """
        Computes the ULP error metrics for the comparison.

        This method calculates three key metrics:
        1. Maximum ULP error: The maximum difference in ULPs between the benchmark and device outputs.
        2. Mean ULP error: The average difference in ULPs between the benchmark and device outputs.
        3. ULP error proportion: The proportion of ULP errors that exceed a certain threshold.

        Args:
        None (this method uses instance variables)

        Returns:
        dict: A dictionary containing the computed ULP error metrics.
            The dictionary has the following keys:
            - "max_ulp_error": The maximum ULP error.
            - "mean_ulp_error": The mean ULP error.
            - "ulp_error_proportion": The proportion of ULP errors exceeding the threshold.
        """
        max_ulp_error = self._stat_max_ulp_err(self.ulp_err)
        mean_ulp_error = self._stat_mean_ulp_err(self.ulp_err)
        
        ulp_error_proportion = self._stat_ulp_error_proportion(self.ulp_err)
        
        return {
            "max_ulp_error": max_ulp_error,
            "mean_ulp_error": mean_ulp_error,
            "ulp_error_proportion": ulp_error_proportion
        }


class UlpPrecisionCompare(BasePrecisionCompare):
    def __init__(self, input_data):
        super().__init__(input_data)
        self.compare_algorithm = CompareConst.ULP_COMPARE_ALGORITHM_NAME

    @staticmethod
    def _compute_ulp_err_proportion_ratio(npu_value, gpu_value, dtype):
        column_name = ApiPrecisionCompareColumn.ULP_ERR_PROPORTION
        if is_inf_or_nan(npu_value) or is_inf_or_nan(gpu_value):
            return check_inf_or_nan(npu_value, gpu_value, column_name)
        else:
            return calc_ratio(npu_value, gpu_value, dtype), True, ""

    def _compute_mean_ulp_err(self):
        column_name = ApiPrecisionCompareColumn.MEAN_ULP_ERR
        npu_value, gpu_value = self._get_and_convert_values(column_name)
        if is_inf_or_nan(npu_value) or is_inf_or_nan(gpu_value):
            _, mean_ulp_err_inf_nan_consistency, message = check_inf_or_nan(npu_value, gpu_value, column_name)
            return npu_value, mean_ulp_err_inf_nan_consistency, message
        else:
            return npu_value, True, ""
    
    def _compute_ulp_err_proportion(self):
        column_name = ApiPrecisionCompareColumn.ULP_ERR_PROPORTION
        npu_value, gpu_value = self._get_and_convert_values(column_name)
        return npu_value, gpu_value
        
    def _get_status(self, metrics, inf_nan_consistency):
        ulp_inf_nan_consistency = inf_nan_consistency.mean_ulp_err_inf_nan_consistency and \
                                  inf_nan_consistency.ulp_err_proportion_ratio_inf_nan_consistency  

        if not ulp_inf_nan_consistency:
            status_dict = {
                CompareConst.ULP_ERR_STATUS: CompareConst.ERROR
            }
            compare_result = CompareConst.ERROR
            metrics[CompareConst.COMPARE_MESSAGE] = metrics.get(CompareConst.COMPARE_MESSAGE, "") + \
                "ERROR: ULP误差不满足标准\n"
            metrics.update({CompareConst.COMPARE_RESULT: compare_result})
            return metrics
        
        dtype = self.row_npu.get(ApiPrecisionCompareColumn.DEVICE_DTYPE)
        mean_ulp_err = metrics.get(CompareConst.MEAN_ULP_ERR)
        ulp_err_proportion = metrics.get(CompareConst.ULP_ERR_PROPORTION)
        ulp_err_proportion_ratio = metrics.get(CompareConst.ULP_ERR_PROPORTION_RATIO)
        if is_dtype_fp8_or_hif8(dtype):
            status, final_message = \
                self._get_fp8_ulp_err_status(ulp_err_proportion)
        elif dtype == Const.TORCH_FLOAT32:
            status, final_message = \
                self._get_fp32_ulp_err_status(mean_ulp_err, ulp_err_proportion, ulp_err_proportion_ratio)
        else:
            status, final_message = \
                self._get_fp16_ulp_err_status(ulp_err_proportion, ulp_err_proportion_ratio)
        metrics[CompareConst.COMPARE_MESSAGE] = metrics.get(CompareConst.COMPARE_MESSAGE, "") + final_message

        status_dict = {
            CompareConst.ULP_ERR_STATUS: status
        }
        compare_result = status
        metrics.update(status_dict)
        metrics.update({CompareConst.COMPARE_RESULT: compare_result})
        return metrics

    def _get_fp32_ulp_err_status(self, mean_ulp_err, ulp_err_proportion, ulp_err_proportion_ratio):
        mean_ulp_err_threshold, ulp_err_proportion_threshold, ulp_err_proportion_ratio_threshold = \
                                                        StandardConfig.get_ulp_threshold(torch.float32)
        if mean_ulp_err < mean_ulp_err_threshold:
            return CompareConst.PASS, ""
        elif ulp_err_proportion < ulp_err_proportion_threshold:
            return CompareConst.PASS, ""
        elif ulp_err_proportion_ratio < ulp_err_proportion_ratio_threshold:
            return CompareConst.PASS, ""
        compare_message = "ERROR: ULP误差不满足标准\n"
        return CompareConst.ERROR, compare_message
        
    def _get_fp16_ulp_err_status(self, ulp_err_proportion, ulp_err_proportion_ratio):
        _, ulp_err_proportion_threshold, ulp_err_proportion_ratio_threshold = \
                                                        StandardConfig.get_ulp_threshold(torch.float16)
        if ulp_err_proportion < ulp_err_proportion_threshold:
            return CompareConst.PASS, ""
        elif ulp_err_proportion_ratio < ulp_err_proportion_ratio_threshold:
            return CompareConst.PASS, ""
        compare_message = "ERROR: ULP误差不满足标准\n"
        return CompareConst.ERROR, compare_message

    def _get_fp8_ulp_err_status(self, ulp_err_proportion):
        _, ulp_err_proportion_threshold, _ = StandardConfig.get_ulp_threshold(torch.float16)
        if ulp_err_proportion < ulp_err_proportion_threshold:
            return CompareConst.PASS, ""
        compare_message = "ERROR: ULP误差不满足标准\n"
        return CompareConst.ERROR, compare_message
    
    def _compute_ratio(self):
        compare_message = ""
        dtype = self.row_npu.get(ApiPrecisionCompareColumn.DEVICE_DTYPE)
        if is_dtype_fp8_or_hif8(dtype):
            mean_ulp_err = CompareConst.SPACE
            ulp_err_proportion_ratio = CompareConst.SPACE
            npu_ulp_err_proportion = self._get_and_convert_value(self.row_npu, 
                                                                 ApiPrecisionCompareColumn.ULP_ERR_PROPORTION, "NPU")
            mean_ulp_err_inf_nan_consistency = True
            ulp_err_proportion_ratio_inf_nan_consistency = True
        else:
            mean_ulp_err, mean_ulp_err_inf_nan_consistency, mean_ulp_err_message = self._compute_mean_ulp_err()
            compare_message += mean_ulp_err_message
            npu_ulp_err_proportion, gpu_ulp_err_proportion = self._compute_ulp_err_proportion()
            ulp_err_proportion_ratio, ulp_err_proportion_ratio_inf_nan_consistency, ulp_err_proportion_ratio_message = \
                self._compute_ulp_err_proportion_ratio(npu_ulp_err_proportion, gpu_ulp_err_proportion, str(self.dtype))
            compare_message += ulp_err_proportion_ratio_message
        metrics = {
            CompareConst.MEAN_ULP_ERR: mean_ulp_err,
            CompareConst.ULP_ERR_PROPORTION: npu_ulp_err_proportion,
            CompareConst.ULP_ERR_PROPORTION_RATIO: ulp_err_proportion_ratio,
            CompareConst.COMPARE_MESSAGE: compare_message
        }
        return metrics, UlpInfNanConsistency(mean_ulp_err_inf_nan_consistency, 
                                             ulp_err_proportion_ratio_inf_nan_consistency)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import math
import torch
import numpy

try:
    import torch_npu
except ImportError:
    IS_GPU = True
else:
    IS_GPU = False

from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check_utils import hf_32_standard_api
from msprobe.pytorch.api_accuracy_checker.common.utils import check_object_type, get_full_data_path, \
    CompareException, get_module_and_atttribute_name, get_attribute, is_dtype_fp8, is_dtype_hif8, is_hifloat8_tensor, \
    is_dtype_fp8_or_hif8
from msprobe.core.common.file_utils import FileChecker, load_npy
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import load_pt
from msprobe.core.common.const import Const, FileCheckConst, CompareConst


TORCH_TYPE = ["torch.device", "torch.dtype"]
TENSOR_DATA_LIST = ["torch.Tensor", "torch.nn.parameter.Parameter"]
FLOAT_TYPE = [
            'torch.float32', 
            'torch.float', 
            'torch.float64', 
            'torch.double', 
            'torch.float16',
            'torch.half', 
            'torch.bfloat16',
            'torch.float8_e4m3fn',
            'torch.float8_e5m2',
            'torch_npu.HiFloat8Tensor'
            ]
NUMPY_TYPE = [
            "numpy.int8", "numpy.int16", "numpy.int32", "numpy.int64", "numpy.uint8", "numpy.uint16", "numpy.uint32",
            "numpy.uint64", "numpy.float16", "numpy.float32", "numpy.float64", "numpy.float128", "numpy.complex64", 
            "numpy.complex128", "numpy.complex256", "numpy.bool_", "numpy.string_", "numpy.bytes_", "numpy.unicode_"
            ]


def gen_data(info, api_name, need_grad, convert_type, real_data_path=None):
    """
    Function Description:
        Based on arg basic information, generate arg data
    Parameter:
        info: arg basic information. Dict
        api_name: API name
        need_grad: set Tensor grad for backward
        convert_type: convert ori_type to dist_type flag.
    """
    check_object_type(info, dict)
    data_type = info.get('type')
    data_path = info.get('datapath', info.get('data_name'))
    data_path = get_full_data_path(data_path, real_data_path)
    dtype = info.get('dtype')
    if is_dtype_fp8_or_hif8(dtype) and IS_GPU:
        raise CompareException("GPU does not need to support float8 data type")
    if data_type in TENSOR_DATA_LIST:
        if data_path:
            data = gen_real_tensor(data_path, convert_type)
        else:
            data = gen_random_tensor(info, convert_type)
        if api_name in hf_32_standard_api and data.dtype == torch.float32:
            data = fp32_to_hf32_to_fp32(data)
        if info.get('requires_grad') and need_grad:
            if is_hifloat8_tensor(data):
                origin_dtype = info.get('dtype')
            else:
                origin_dtype = data.dtype
            if is_dtype_fp8(origin_dtype):
                data = data.to(torch.float32)
            data.requires_grad_(True)
            temp_data = data * 1
            data = temp_data.type_as(data)
            data.retain_grad()
            if is_dtype_fp8(origin_dtype):
                data = data.to(origin_dtype)
            if is_dtype_hif8(origin_dtype):
                data = torch_npu.HiFloat8Tensor.to_hifloat8(data)
    elif data_type.startswith("numpy"):
        if data_type not in NUMPY_TYPE:
            raise Exception("{} is not supported now".format(data_type))
        data = info.get("value")
        try:
            module_name, attribute_name = get_module_and_atttribute_name(data_type)
            data = get_attribute(module_name, attribute_name)(data)
        except Exception as err:
            logger.error("Failed to convert the type to numpy: %s" % str(err))
    elif data_type == "torch.Size":
        data = torch.Size(info.get("value"))
    else:
        data = info.get('value')
        if info.get("type") == "slice":
            data = slice(*data)
        if info.get("type") == "ellipsis":
            data = ...
    return data


def gen_real_tensor(data_path, convert_type):
    """
    Function Description:
        Based on API data path, generate input parameters real data
    Parameter:
        data_path: API data path
        convert_type: convert ori_type to dist_type flag.
    """
    data_path = os.path.realpath(data_path)
    data_path_checker = FileChecker(data_path, FileCheckConst.FILE, ability=FileCheckConst.READ_ABLE)
    data_path = data_path_checker.common_check()
    if not data_path.endswith('.pt') and not data_path.endswith('.npy'):
        error_info = f"The file: {data_path} is not a pt or numpy file."
        raise CompareException(CompareException.INVALID_FILE_ERROR, error_info)
    if data_path.endswith('.pt'):
        data = load_pt(data_path, to_cpu=True)
    else:
        data_np = load_npy(data_path)
        data = torch.from_numpy(data_np)
    if convert_type:
        ori_dtype = Const.CONVERT.get(convert_type)[0]
        dist_dtype = Const.CONVERT.get(convert_type)[1]
        module_name, attribute_name = get_module_and_atttribute_name(dist_dtype)
        if str(data.dtype) == ori_dtype:
            data = data.type(get_attribute(module_name, attribute_name))
    return data


def gen_random_tensor(info, convert_type):
    """
    Function Description:
        Based on API MAX and MIN, generate input parameters random data
    Parameter:
        info: API data info
        convert_type: convert ori_type to dist_type flag.
    """
    check_object_type(info, dict)

    low_origin = info.get('Min')
    low = info.get('Min_except_inf_nan', low_origin)
    high_origin = info.get('Max')
    high = info.get('Max_except_inf_nan', high_origin)
    
    low_info = [low, low_origin]
    high_info = [high, high_origin]
    data_dtype = info.get('dtype')
    shape = tuple(info.get('shape'))
    if 0 in shape:
        low, low_origin = 0, 0
        high, high_origin = 0, 0
        low_info = [low, low_origin]
        high_info = [high, high_origin]
    elif not isinstance(low, (int, float)) or not isinstance(high, (int, float)):
        error_info = f'Data info Min: {low} , Max: {high}, info type must be int or float.'
        raise CompareException(CompareException.INVALID_PARAM_ERROR, error_info)
    if data_dtype == "torch.bool":
        data = gen_bool_tensor(low, high, shape)
    else:
        data = gen_common_tensor(low_info, high_info, shape, data_dtype, convert_type)
    return data


def fp32_to_hf32_to_fp32(input_tensor):
    # 将输入的float32 tensor转为hf32 tensor，再转为float32 tensor
    input_np = input_tensor.detach().numpy()
    input_int = input_np.view(numpy.int32)
    input_int = numpy.right_shift(numpy.right_shift(input_int, 11) + 1, 1)
    input_int = numpy.left_shift(input_int, 12)
    input_fp32 = input_int.view(numpy.float32)
    input_hf32 = torch.from_numpy(input_fp32)
    return input_hf32


def gen_common_tensor(low_info, high_info, shape, data_dtype, convert_type):
    """
    Function Description:
        Based on API basic information, generate int or float tensor
    Parameter:
        low_info: [low, low_origin], low is the minimum value in the tensor removed inf and nan, 
        low_origin is the original minimum value in the tensor
        high_info: [high, high_origin], high is the maximum value in the tensor removed inf and nan, 
        high_origin is the original maximum value in the tensor
        shape:The shape of Tensor
        data_dtype: The data type of Tensor
        convert_type: convert ori_type to dist_type flag.
    """
    if convert_type:
        ori_dtype = Const.CONVERT.get(convert_type)[0]
        if ori_dtype == data_dtype:
            data_dtype = Const.CONVERT.get(convert_type)[1]
    low, low_origin = low_info[0], low_info[1]
    high, high_origin = high_info[0], high_info[1]
    module_name, attribute_name = get_module_and_atttribute_name(data_dtype)
    dtype = get_attribute(module_name, attribute_name)
    if data_dtype in FLOAT_TYPE: 
        if math.isnan(high):
            tensor = torch.full(shape, float('nan'), dtype=dtype)
            return tensor
        #high_origin为新版json中的属性，只有当high_origin不为None,且high为inf或-inf时，原tensor全为inf或-inf
        if high_origin and high in [float(CompareConst.INF), float(CompareConst.NEG_INF)]:
            tensor = torch.full(shape, high, dtype=dtype)
            tensor[-1] = low
            return tensor
        
        low_scale, high_scale = low, high
        if is_dtype_hif8(dtype):
            finfo_dtype = torch.float32
        else:
            finfo_dtype = dtype
        dtype_finfo = torch.finfo(finfo_dtype)
        #适配老版json high和low为inf或-inf的情况，取dtype的最大值或最小值进行放缩
        if high == float(CompareConst.INF):
            high_scale = dtype_finfo.max
        elif high == float(CompareConst.NEG_INF):
            high_scale = dtype_finfo.min
        if low == float(CompareConst.INF):
            low_scale = dtype_finfo.max
        elif low == float(CompareConst.NEG_INF):
            low_scale = dtype_finfo.min

        scale = high_scale - low_scale
        if is_dtype_fp8(dtype) or is_dtype_hif8(dtype):
            generate_dtype = torch.float32
        else:
            generate_dtype = dtype
        rand01 = torch.rand(shape, dtype=generate_dtype)
        tensor = rand01 * scale + low_scale
    elif 'int' in data_dtype or 'long' in data_dtype:
        low, high = int(low), int(high)
        tensor = torch.randint(low, high + 1, shape, dtype=dtype)
    else:
        logger.error('Dtype is not supported: ' + data_dtype)
        raise NotImplementedError()
    if tensor.nelement() == 0:
        return tensor
    tmp_tensor = tensor.reshape(-1)
    if high_origin and math.isnan(high_origin):
        if tmp_tensor.numel() <= 2:
            tmp_tensor[0] = float('nan')
            tmp_tensor[-1] = high
        else:
            tmp_tensor[0] = low
            tmp_tensor[1] = float('nan')
            tmp_tensor[-1] = high
    else:
        tmp_tensor[0] = low
        tmp_tensor[-1] = high
        if high_origin in [float(CompareConst.INF), float(CompareConst.NEG_INF)]:
            tmp_tensor[-1] = high_origin
        if low_origin in [float(CompareConst.INF), float(CompareConst.NEG_INF)]:
            tmp_tensor[0] = low_origin
    data = tmp_tensor.reshape(shape)
    if is_dtype_fp8(dtype):
        data = data.to(dtype)
    if is_dtype_hif8(dtype):
        if IS_GPU:
            raise CompareException("GPU does not support torch_npu.HiFloat8Tensor")
        data = torch_npu.HiFloat8Tensor.to_hifloat8(data)
    return data


def gen_bool_tensor(low, high, shape):
    """
    Function Description:
        Based on API basic information, generate bool tensor
    Parameter:
        low: The minimum value in Tensor
        high: The max value in Tensor
        shape:The shape of Tensor
    """
    low, high = int(low), int(high)
    if low > high:
        low, high = high, low
    tensor = torch.randint(low, high + 1, shape)
    data = torch.gt(tensor, 0)
    return data


def gen_args(args_info, api_name, func_options):
    """
    Function Description:
        Based on API basic information, generate input parameters: args, for API forward running
    Parameter:
        api_info: API basic information. List
        api_name: API name
        need_grad: set Tensor grad for backward
        convert_type: convert ori_type to dist_type flag.
        real_data_path: the root directory for storing real data.
    """
    check_object_type(args_info, list)
    args_result = []
    
    need_grad = func_options.get('need_grad', True)
    convert_type = func_options.get('convert_type', None)
    real_data_path = func_options.get('real_data_path', None)
    depth = func_options.get('depth', 0)

    if depth > Const.MAX_DEPTH:
        logger.error("The depth of args is too large, please check the input args.")
        raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
    
    for arg in args_info:
        if isinstance(arg, (list, tuple)):
            func_options['depth'] = depth + 1
            data = gen_args(arg, api_name, func_options)
        elif isinstance(arg, dict):
            data = gen_data(arg, api_name, need_grad, convert_type, real_data_path)
        elif arg is None:
            data = None
        else:
            logger.warning(f'Warning: {arg} is not supported')
            raise NotImplementedError()
        args_result.append(data)
    return args_result


def gen_kwargs(api_info, api_name, convert_type=None, real_data_path=None):
    """
    Function Description:
        Based on API basic information, generate input parameters: kwargs, for API forward running
    Parameter:
        api_info: API basic information. Dict
        api_name: API name
        convert_type: convert ori_type to dist_type flag.
        real_data_path: the root directory for storing real data.
    """
    check_object_type(api_info, dict)
    kwargs_params = api_info.get("input_kwargs")
    for key, value in kwargs_params.items():
        if isinstance(value, (list, tuple)):
            kwargs_params[key] = gen_list_kwargs(value, api_name, convert_type, real_data_path)
        elif value is None:
            kwargs_params[key] = None
        elif key == 'atten_mask' and api_name == 'npu_fusion_attention':
            sparse_mode = kwargs_params.get('sparse_mode', {})
            if isinstance(sparse_mode, dict):
                sparse_mode_value = sparse_mode.get('value', 0)
            elif isinstance(sparse_mode, int):
                sparse_mode_value = sparse_mode
            else:
                msg = f'The sparse_mode value is not int or dict, but {type(sparse_mode)}'
                raise CompareException(CompareException.INVALID_PARAM_ERROR, msg)
            if sparse_mode_value in Const.FA_SPECIAL_SPARSE_MODE:
                kwargs_params[key] = gen_atten_mask(value, convert_type, real_data_path)
            else:
                kwargs_params[key] = gen_data(value, api_name, True, convert_type, real_data_path)
        elif value.get('type') in TENSOR_DATA_LIST or value.get('type').startswith("numpy"):
            kwargs_params[key] = gen_data(value, api_name, True, convert_type, real_data_path)
        elif value.get('type') in TORCH_TYPE:
            gen_torch_kwargs(kwargs_params, key, value)
        else:
            kwargs_params[key] = value.get('value')
    return kwargs_params


def gen_atten_mask(info, convert_type, real_data_path):
    """
    Function Description:
        Based on API basic information, generate input parameters: atten_mask, for API forward running
    Parameter:
        info: API basic information. Dict
        convert_type: convert ori_type to dist_type flag.
        real_data_path: the root directory for storing real data.
    """
    check_object_type(info, dict)
    data_type = info.get('type')
    data_path = info.get('datapath', info.get('data_name'))
    data_path = get_full_data_path(data_path, real_data_path)
    data = None
    if data_type in TENSOR_DATA_LIST:
        if data_path:
            data = gen_real_tensor(data_path, convert_type)
        else:
            # 生成一个2048x2048的三角矩阵，对角线为1，其余为0
            # 这是npu_fusion_attention的sparse_mode为[2, 3, 4]时，atten_mask的shape
            data = torch.triu(torch.ones([2048, 2048]), diagonal=1).to(torch.bool)
    return data


def gen_torch_kwargs(kwargs_params, key, value):
    if value.get('type') != "torch.device":
        module_name, attribute_name = get_module_and_atttribute_name(value.get('value'))
        kwargs_params[key] = get_attribute(module_name, attribute_name)


def gen_list_kwargs(kwargs_item_value, api_name, convert_type, real_data_path=None):
    """
    Function Description:
        When kwargs value is list, generate the list of kwargs result
    Parameter:
        kwargs_item_value: kwargs value before to generate. List
        api_name: API name
        convert_type: convert ori_type to dist_type flag.
    """
    kwargs_item_result = []
    for item in kwargs_item_value:
        if item.get('type') in TENSOR_DATA_LIST:
            item_value = gen_data(item, api_name, False, convert_type, real_data_path)
        elif item.get('type') == "torch.Size":
            item_value = torch.Size(item.get('value'))
        else:
            item_value = item.get('value')
        kwargs_item_result.append(item_value)
    return kwargs_item_result


def get_output_dtype(api_info):
    """
    Function Description:
        Based on API basic information, get the output data dtype
    Parameter:
        api_info: API basic information. Dict
    """
    output_dtype = None
    output_info = api_info.get(Const.OUTPUT)
    if output_info and isinstance(output_info[0], dict):
        output_str_dtype = output_info[0].get(Const.DTYPE)
        if output_str_dtype in Const.TORCH_FLOAT_DTYPE:
            module_name, attribute_name = get_module_and_atttribute_name(output_str_dtype)
            output_dtype = get_attribute(module_name, attribute_name)
    return output_dtype


def gen_api_params(api_info, api_name, need_grad=True, convert_type=None, real_data_path=None):
    """
    Function Description:
        Based on API basic information, generate input parameters: args, kwargs, for API forward running
    Parameter:
        api_info: API basic information. Dict
        api_name: API name
        need_grad: set grad for backward
        convert_type: convert ori_type to dist_type flag.
    """
    check_object_type(api_info, dict)
    if convert_type and convert_type not in Const.CONVERT:
        error_info = f"convert_type params not support {convert_type}."
        raise CompareException(CompareException.INVALID_PARAM_ERROR, error_info)
    kwargs_params = gen_kwargs(api_info, api_name, convert_type, real_data_path)
    func_options = {
        'need_grad': need_grad,
        'convert_type': convert_type,
        'real_data_path': real_data_path,
        'depth': 0
    }
    if api_info.get("input_args"):
        args_params = gen_args(api_info.get("input_args"), api_name, func_options)
    else:
        logger.warning(f'Warning: No args in {api_info} ')
        args_params = []
    output_dtype = get_output_dtype(api_info)
    return args_params, kwargs_params, output_dtype



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import sys

import numpy as np

from msprobe.core.common.log import logger
from msprobe.infer.utils.check.string_checker import StringChecker
from msprobe.infer.utils.file_open_check import ms_open, MAX_SIZE_LIMITE_NORMAL_FILE


IS_MSACCUCMP_PATH_SET = False
GLOBAL_TENSOR_CONVERTER = None


def _lazy_import_torch():
    try:
        import torch  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as err:
        if err.name == "torch":
            logger.error("Missing dependency: torch. Please install torch to use torchair compare.")
        raise
    return torch


def default_tensor_converter(tensor):
    torch = _lazy_import_torch()
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"Expected a torch.Tensor, but got {type(tensor).__name__}")
    return tensor.data.reshape(tensor.shape)


def set_msaccucmp_path_from_cann():
    global IS_MSACCUCMP_PATH_SET
    global GLOBAL_TENSOR_CONVERTER

    # env TOOLCHAIN_HOME works for both development and product packages.
    cann_path = os.environ.get("TOOLCHAIN_HOME", os.environ.get("ASCEND_TOOLKIT_HOME", ""))
    if not cann_path:
        raise OSError("CANN toolkit in not installed or not set, try installing the latest CANN toolkit.")
    if StringChecker().is_str_safe().check(cann_path, True):
        cann_path = cann_path.split(":")[0]  # Could be multiple split by :, should use the first one

    msaccucmp_path = os.path.join(cann_path, "tools", "operator_cmp", "compare")
    if not os.path.exists(msaccucmp_path):
        raise OSError(f"{msaccucmp_path} not exists, try installing the latest CANN toolkit.")

    if msaccucmp_path not in sys.path:
        sys.path.append(msaccucmp_path)
    IS_MSACCUCMP_PATH_SET = True
    logger.info(f"Set msaccucmp_path={msaccucmp_path}")

    if GLOBAL_TENSOR_CONVERTER is None:
        from conversion import tensor_conversion

        if hasattr(tensor_conversion, "ConvertSingleTensorFormat"):
            GLOBAL_TENSOR_CONVERTER = tensor_conversion.ConvertSingleTensorFormat()
        else:
            GLOBAL_TENSOR_CONVERTER = default_tensor_converter
            logger.warning("ConvertSingleTensorFormat not found in msaccucmp, connot convert tensor format."
                           " Try installing the latest CANN toolkit."
                           )


def parse_torchair_dump_data(dump_file):
    if dump_file.endswith(".npz"):  # Custom converted data info
        with ms_open(dump_file, "rb", max_size=MAX_SIZE_LIMITE_NORMAL_FILE) as f:
            loaded = np.load(f)
        return loaded.get("inputs", []), loaded.get("outputs", [])

    if not IS_MSACCUCMP_PATH_SET:
        set_msaccucmp_path_from_cann()
    from dump_parse.dump_utils import parse_dump_file  # Parser tool from CANN msaccucmp
    from cmp_utils.constant.const_manager import ConstManager

    bin_dump_data = parse_dump_file(dump_file, dump_version=ConstManager.OLD_DUMP_TYPE)
    inputs = [GLOBAL_TENSOR_CONVERTER(input_data) for input_data in bin_dump_data.input_data]
    outputs = [GLOBAL_TENSOR_CONVERTER(output_data) for output_data in bin_dump_data.output_data]
    return inputs, outputs



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves xxxx function.
"""

import numpy as np

from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse import dump_utils


class OverflowDetection:
    """
    The class for op overflow detection info
    """

    def __init__(self: any, *args: any) -> None:
        compare_data, op_name = args
        self.dump_version = compare_data.dump_version
        self.left_dump_info = compare_data.left_dump_info
        self.overflow_tensor_list = []
        self.op_name = op_name

    @staticmethod
    def process_model_overflow_detection(op_name: str, index: int, is_input: bool, tensor: any) -> str:
        """
        process model overflow detection
        """
        if tensor and tensor.data_type == DD.DT_FLOAT16:
            tensor_data_array = tensor.data
            overflow_result = OverflowDetection._judge_overflow_data_by_array(tensor_data_array)
            if overflow_result == 'YES':
                log.print_warn_log(
                    "{} operator {}:{} is overflow.".format(op_name, 'input' if is_input else 'output', index))
            return overflow_result
        return ConstManager.NAN

    @staticmethod
    def _judge_overflow_data_by_array(tensor_data: any) -> str:
        if tensor_data is not None:
            absolute_value_data = np.absolute(tensor_data)
            up_overflow_flag = (absolute_value_data.max() >= ConstManager.OVERFLOW_MAX_VALUE)
            if up_overflow_flag:
                return 'YES'
            return 'NO'
        return ConstManager.NAN

    def process_op_overflow_detection(self: any) -> None:
        """
        process op overflow detection
        """
        log.print_info_log("Checking {} operator overflow".format(self.op_name))
        input_tensor_data_info, output_tensor_data_info = self.parse_dump_file()
        if len(input_tensor_data_info) == 0 and len(output_tensor_data_info) == 0:
            return
        self._check_overflow_tensor(input_tensor_data_info, output_tensor_data_info)
        self._print_overflow_info_to_console(self.overflow_tensor_list)

    def parse_dump_file(self: any) -> (list, list):
        """
        process op overflow detection
        """
        dump_file_list, dump_mode = self.left_dump_info.get_op_dump_file(self.op_name)
        if not dump_file_list:
            log.print_error_log("dump_file_list is empty, please check.")
            raise CompareError(CompareError.MSACCUCMP_MISSING_FILE)
        dump_file_path = dump_file_list[-1]
        if not dump_file_path:
            return [], []
        dump_data = dump_utils.parse_dump_file(dump_file_path, self.dump_version)
        input_tensor_data_info = self._get_tensor_data_info("input", dump_data.input_data, dump_file_path)
        output_tensor_data_info = self._get_tensor_data_info("output", dump_data.output_data, dump_file_path)
        return input_tensor_data_info, output_tensor_data_info

    def _get_tensor_data_info(self: any, tensor_type: str, tensor_list: list, dump_file_path: str) -> list:
        tensor_data_info = []
        for (index, tensor) in enumerate(tensor_list):
            if tensor and tensor.data_type == DD.DT_FLOAT16:
                log.print_info_log('Start to parse the data of %s:%d in "%r".' % (tensor_type, index, dump_file_path))
                array = tensor.data
                tensor_data_info.append(
                    {"tensor_type": tensor_type, "index": str(index), "tensor_data": array, "tensor_info": tensor,
                     "dump_file_path": dump_file_path})
        if len(tensor_data_info) == 0:
            log.print_warn_log("The {} data type of {} operator is not float16. "
                               "The overflow check supports only float16.".format(tensor_type, self.op_name))
        return tensor_data_info

    def _check_overflow_tensor(self: any, input_data: list, output_data: list) -> None:
        self._judge_overflow_data(input_data, "input")
        self._judge_overflow_data(output_data, "output")

    def _judge_overflow_data(self: any, tensor_list: list, tensor_type: str) -> None:
        if len(tensor_list) == 0:
            dump_file_list, dump_mode = self.left_dump_info.get_op_dump_file(self.op_name)
            dump_file_path = dump_file_list[-1]
            log.print_warn_log('There is no %s in "%r".' % (
                tensor_type, dump_file_path))
            return
        for item in tensor_list:
            tensor_data = item.get("tensor_data")
            absolute_value_data = np.absolute(tensor_data)
            up_overflow_flag = (absolute_value_data.max() >= ConstManager.OVERFLOW_MAX_VALUE)
            if up_overflow_flag:
                self.overflow_tensor_list.append(item)

    def _print_overflow_info_to_console(self: any, overflow_tensor_list: list) -> None:
        if len(overflow_tensor_list) == 0:
            log.print_info_log("The input and output of the op: {} is not overflow.".format(self.op_name))
        else:
            tensor_index_len = []
            tensor_index_info = []
            for item in overflow_tensor_list:
                tensor_index = "".join([item.get("tensor_type"), ":", item.get("index")])
                tensor_index_info.append(tensor_index)
                tensor_index_len.append(len(tensor_index))
            spacing_distance = max(tensor_index_len)
            if spacing_distance < 20:
                spacing_distance = 20
            format_line = "".join(['{:<', str(spacing_distance), '}', '{:<', str(spacing_distance), '}'])
            log.print_info_log(format_line.format("TensorIndex", "Overflow"))
            for item in tensor_index_info:
                log.print_info_log(format_line.format(item, "yes"))
            log.print_info_log("%s operator has overflowed." % self.op_name)



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the dump data proto function.
"""

import numpy as np

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


def contain_depth_dimension(tensor_format: any) -> bool:
    """
    Contain depth dimension
    :param tensor_format: the tensor format
    :return: bool, if true contain depth dimension
    """
    return tensor_format in [DD.FORMAT_NDHWC, DD.FORMAT_NCDHW, DD.FORMAT_DHWCN, DD.FORMAT_NDC1HWC0]


def get_format_string(dump_format: any) -> str:
    """
    Get format string by dump_format
    :param dump_format: the dump format
    :return: the string
    """

    for key, value in ConstManager.STRING_TO_FORMAT_MAP.items():
        if dump_format == value:
            return key
    message = "The format({}) does not support.".format(str(dump_format))
    log.print_error_log(message)
    raise CompareError(CompareError.MSACCUCMP_INVALID_FORMAT_ERROR, message)


def get_data_type_by_dtype(dtype: any) -> any:
    """
    Get data type by dtype
    :param dtype: the numpy dtype
    :return:the OutputDataType
    """
    if dtype == np.double:
        return DD.DT_DOUBLE
    for (key, value) in list(ConstManager.DATA_TYPE_TO_DTYPE_MAP.items()):
        if dtype == value.get(ConstManager.DTYPE_KEY):
            return key
    message = "The dtype({}) does not support.".format(str(dtype))
    log.print_error_log(message)
    raise CompareError(CompareError.MSACCUCMP_INVALID_DATA_TYPE_ERROR, message)


def get_dtype_by_data_type(data_type: any) -> any:
    """
    Get dtype by output data type
    :param data_type: OutputDataType
    :return: the dtype
    """
    if data_type not in ConstManager.DATA_TYPE_TO_DTYPE_MAP:
        message = "The output data type ({}) does not support." .format(str(data_type))
        log.print_error_log(message)
        raise CompareError(CompareError.MSACCUCMP_INVALID_DATA_TYPE_ERROR, message)
    dtype = ConstManager.DATA_TYPE_TO_DTYPE_MAP.get(data_type).get(ConstManager.DTYPE_KEY)
    if dtype == 'bfloat16':
        try:
            from ml_dtypes import bfloat16
        except ModuleNotFoundError as ee:
            raise TypeError('bfloat16 is not supported in numpy, run `pip install ml_dtypes` for support.') from ee
        dtype = bfloat16
    return dtype


def get_struct_format_by_data_type(data_type: any) -> any:
    """
    Get the struct format for data type
    :param data_type: the data type
    :return: struct format
    """
    if data_type not in ConstManager.DATA_TYPE_TO_DTYPE_MAP:
        message = "The output data type({}) does not support.".format(str(data_type))
        log.print_error_log(message)
        raise CompareError(CompareError.MSACCUCMP_INVALID_DATA_TYPE_ERROR, message)
    return ConstManager.DATA_TYPE_TO_DTYPE_MAP.get(data_type).get(ConstManager.STRUCT_FORMAT_KEY)


def get_sub_format(tensor: any, default_value: int = 0) -> int:
    """
    Adapt group conv sub_format, to get group
    :param tensor: the tensor
    :param default_value: default value for sub format
    :return: the value of sub format
    """
    sub_format = default_value
    if hasattr(tensor, 'sub_format'):
        sub_format = tensor.sub_format
    return sub_format



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the common function.
"""
import os
import re
import math
import collections
import csv
import numpy as np

from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.utils_type import ShapeType
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


PATH_BLACK_LIST_REGEX = re.compile(r"[^_A-Za-z0-9/.,-]")  # Includes `,`
MALICIOUS_CSV_PATTERN = re.compile(r'^[＝＋－\+\-=%@]|;[＝＋－\+\-=%@]')


def safe_path_string(value):
    if re.search(PATH_BLACK_LIST_REGEX, value):
        raise ValueError("String parameter contains invalid characters.")
    return value


def sanitize_csv_value(value: str, errors='strict'):
    if errors == 'ignore' or not isinstance(value, str):
        return value

    sanitized_value = value
    try:
        float(value) # in case value is a digit but in str format
    except ValueError as e: # not digit
        if not MALICIOUS_CSV_PATTERN.search(value):
            pass
        elif errors == 'replace':
            sanitized_value = ' ' + value
        else:
            msg = f'Malicious value is not allowed to be written to the csv: {value}'
            log.print_error_log("Please check the value written to the csv")
            raise ValueError(msg) from e

    return sanitized_value


def make_msnpy_file_name(file_path: str, op_name: str, tensor_type: str, index: int, tensor_format: int) -> str:
    """
    Make file name for msnpy
    :param file_path: the file path
    :param op_name: the op name
    :param tensor_type: the tensor type, input or output
    :param index: the tensor index
    :param tensor_format: the tensor format
    :return: the msnpy file name
    """
    name_split = os.path.basename(file_path).split('.')
    if len(name_split) == ConstManager.OFFLINE_FILE_NAME_COUNT and op_name:
        # the index 1 for op_name
        name_split[1] = op_name.split('/')[-1]
        origin_file_name = ".".join(name_split)
    else:
        origin_file_name = os.path.basename(file_path)
    return '%s.%s.%d.%s.npy' % (origin_file_name, tensor_type, index, common.get_format_string(tensor_format))


def check_shape_valid_in_nz(shape: list, tensor_shape: list, is_convert_mode: bool = True) -> None:
    """
    check fractal nz dump data shape is valid
    param:
        shape: target shape
        tensor_shape: current tensor shape
        is_convert_shape: the method is used for two mode, one is compare mode, the other is convert mode
    return: None
    """
    if len(shape) == 0:
        error_msg = 'The format before transfer is FRACTAL_NZ. Please enter a valid shape.'
        _raise_exception_by_convert_mode(is_convert_mode, error_msg)
    origin_shape = []
    for index in range(len(tensor_shape) - 4):
        origin_shape.append(tensor_shape[index])
    origin_shape.append(tensor_shape[-2] * tensor_shape[-3])
    origin_shape.append(tensor_shape[-1] * tensor_shape[-4])
    is_invalid_shape = shape[-1] > origin_shape[-1] or \
                     shape[-1] <= origin_shape[-1] - 16 or \
                     shape[-2] > origin_shape[-2] or \
                     shape[-2] <= origin_shape[-2] - 16
    if len(shape) != len(origin_shape) or is_invalid_shape:
        error_msg = 'The target shape %s is invalid. The recommended shape is %s.' \
            % (convert_shape_to_string(shape), convert_shape_to_string(origin_shape))
        _raise_exception_by_convert_mode(is_convert_mode, error_msg)
    for index in range(len(origin_shape) - 2):
        if shape[index] != origin_shape[index]:
            error_msg = 'The target shape %s is invalid, the recommended shape is %s.' \
                % (convert_shape_to_string(shape), convert_shape_to_string(origin_shape))
            _raise_exception_by_convert_mode(is_convert_mode, error_msg)


def get_string_from_list(string_list: list, splitter: str = ',') -> str:
    """
    Get string from list splitter by splitter
    :param string_list: the list to string
    :param splitter: the splitter, default is ','
    :return: the string
    """
    list_str = []
    for item in string_list:
        if isinstance(item, str):
            list_str.append(item)
        else:
            list_str.append(str(item))
    return splitter.join(list_str)


def convert_ndarray_to_bytes(array: np.ndarray) -> bytes:
    """
    convert ndarray to bytes
    @param array: ndarray
    @return:bytes
    """
    return array.tobytes()


def convert_shape_to_string(shape: list) -> str:
    """
    Convert shape to string
    :param shape: the shape
    :return: the shape string
    """
    return "(%s)" % get_string_from_list(shape, ', ')


def format_value(value: float) -> str:
    """
    Format value, 6 decimal places
    :param value: the value to format
    :return: value with 6 decimal places
    """
    return '{:.6f}'.format(value)


def space_to_comma(value: str) -> str:
    """
    Format convert(space to comma)
    :param value: the value to convert
    :return: the value after convert
    """
    new_value = value.replace(',', '|')
    new_value = new_value.replace(' ', ',')
    return new_value.replace('|', ' ')


def merge_dict(dict_dst: dict, dict_src: dict) -> None:
    """
    Merge dict2 into dict1
    :param dict_dst:
    :param dict_src:
    """
    for key in dict_src.keys():
        if key in dict_dst:
            dict_dst[key] = dict_dst[key] + dict_src[key]
        else:
            dict_dst[key] = dict_src[key]


def sort_result_file_by_index(result_file: str, csv_file: bool = True) -> None:
    """
    Sort compare result
    :param result_file: output file path
    :param csv_file: the result is csv or not
    """
    try:
        # read result file and sort result
        if result_file:
            _sort_result_file_exec(result_file, csv_file)
    except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
        log.print_open_file_error(result_file, error)


def get_shape_type(shape_dim_array: list) -> ShapeType:
    """
    Get shape type
    :param shape_dim_array: the shape info
    :return: ShapeType
    """
    return ShapeType.Scalar if sum(shape_dim_array) == len(shape_dim_array) else ShapeType.Tensor


def get_data_type(dump_data_type: str) -> str:
    """
    Get data type
    :param dump_data_type: the shape info
    :return: data type
    """
    if dump_data_type not in ConstManager.DATA_TYPE_TO_STR_DTYPE_MAP:
        return ConstManager.NAN
    return ConstManager.DATA_TYPE_TO_STR_DTYPE_MAP.get(dump_data_type)


def get_address_from_tensor(tensor: any):
    """
    get address from tensor
    args:tensor
    return:address
    """
    if hasattr(tensor, "address") and tensor.address != 0:
        return tensor.address
    else:
        return ConstManager.NAN


def dump_path_contains_npy(dump_path: str) -> bool:
    """
    check dump_file is npy file in dump path
    args: dump_path
    returns: bool
    """
    if dump_path and os.path.isfile(dump_path):
        return dump_path.endswith(ConstManager.NUMPY_SUFFIX)
    if dump_path and os.path.isdir(dump_path):
        return has_npy_at_dir(dump_path)
    return False


def has_npy_at_dir(dump_path: str) -> bool:
    """
    check there is npy file at dump_path
    args:dump_path
    return:bool
    """
    file_list = os.listdir(dump_path)
    for file_path in file_list:
        if str(file_path).endswith(ConstManager.NUMPY_SUFFIX):
            return True
    return False


def _raise_exception_by_convert_mode(is_convert_mode: bool, error_msg: str):
    if is_convert_mode:
        log.print_invalid_nz_dump_data(error_msg, is_error=True)
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, error_msg)
    else:
        raise CompareError(CompareError.MSACCUCMP_INVALID_FRACTAL_NZ_DUMP_DATA_ERROR, error_msg)


def _write_sorted_result(result_file: str, sorted_result_line: list, header_list: list, table_header_info: str,
                         csv_file: bool) -> None:
    with os.fdopen(os.open(result_file, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES), 'w',
                   newline="") as fp_write:
        for item in sorted_result_line:
            if len(item) < 2:
                log.print_error_log('Failed to write sorted result')
                raise IndexError("Invalid data to save, please check it!")
        if csv_file:
            # write header to file
            writer = csv.writer(fp_write)
            writer.writerow((sanitize_csv_value(ii) for ii in header_list))
            # write sorted result to file
            for item in sorted_result_line:
                writer.writerow((sanitize_csv_value(ii) for ii in item[1]))
        else:
            # write header to file
            fp_write.write(table_header_info)
            # write value to file
            for item in sorted_result_line:
                fp_write.write(item[1])


def _get_header_and_data(csv_file: bool, fp_read: any) -> (str, list, list):
    table_header_info = next(fp_read)
    header_list = []
    origin_result_line = []
    if csv_file:
        header_list = table_header_info.strip().split(',')
        result_reader = csv.reader(fp_read)
        for line in result_reader:
            origin_result_line.append((int(line[0]), line))
    else:
        result_reader = fp_read.readlines()
        for line in result_reader:
            origin_result_line.append((int(line.split(" ")[0]), line))
    return table_header_info, header_list, origin_result_line


def _sort_result_file_exec(result_file: str, csv_file: bool = True) -> None:
    check_file_size(result_file, ConstManager.ONE_HUNDRED_MB)
    with open(result_file, 'r') as fp_read:
        table_header_info, header_list, origin_result_line = _get_header_and_data(csv_file, fp_read)
        sorted_result_line = sorted(origin_result_line, key=lambda s: s[0])
    _write_sorted_result(result_file, sorted_result_line, header_list, table_header_info, csv_file)


def check_file_size(file_path: str, size_limit: int, is_raise=False) -> None:
    try:
        file_size = os.path.getsize(file_path)
    except OSError as os_error:
        log.print_open_file_error(file_path, os_error)
        raise CompareError(CompareError.MSACCUCMP_OPEN_FILE_ERROR) from os_error
    if file_size > size_limit:
        log.print_warn_log(
            'The size (%d) of %r exceeds %dMB, it may task more time to run, please wait.'
            % (file_size, file_path, size_limit / 1024 / 1024))
        if is_raise:
            raise CompareError("%r file size (%d) exceeds %d" % (file_path, file_size, size_limit))


def least_common_multiple(left: int, right: int) -> int:
    """
    Least common multiple, in this file, n could not zero
    :param left: One of the calculation parameters
    :param right: One of the calculation parameters
    :return: left, right Least common multiple
    """
    if left == 0 or right == 0:
        return 0
    return (left * right) // math.gcd(left, right)


def ceiling_divide(left: int, right: int) -> int:
    """
    Ceiling divide, in this file, n could not zero
    :param left: One of the calculation parameters
    :param right: One of the calculation parameters
    :return: left, right Ceiling divide
    """
    if right == 0:
        raise ZeroDivisionError('Can not divide zero.')
    return (left + right - 1) // right


ResultInfo = collections.namedtuple(
    "ResultInfo",
    ["op_name", "dump_match", "result_list",
     "ret", "input_list", "input_result_list",
     "output_result_list", "is_ffts",
     "op_name_origin_output_index_map", "npu_vs_npu"])




# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
CosineSimilarity algorithm. This file mainly involves the compare function.
"""

import numpy as np

from msprobe.msaccucmp.algorithm_manager.algorithm_parameter import AlgorithmParameter
from msprobe.msaccucmp.cmp_utils import utils, utils_type, log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager


def compare(my_output_dump_data: any, ground_truth_dump_data: any, args: AlgorithmParameter) -> (str, str):
    """
    compare my output dump data and the ground truth dump data
    by cosine similarity
    cos(sitar) = sum(x[i] * y[i]) /
    (sqrt(sum(x[i] * x[i])) * sqrt(sum(y[i] * y[i])))
    :param my_output_dump_data: my output dump data to compare
    :param ground_truth_dump_data: the ground truth dump data to compare
    :param args: the algorithm parameter
    :return: the result of cosine similarity value and error message (the default is "")
    """
    if args.shape_type == utils_type.ShapeType.Scalar:
        return utils.format_value(1), "This tensor is scalar."

    my_output_norm = np.linalg.norm(my_output_dump_data, axis=-1, keepdims=True)
    ground_truth_norm = np.linalg.norm(ground_truth_dump_data, axis=-1, keepdims=True)
    if my_output_norm <= ConstManager.FLOAT_EPSILON and ground_truth_norm < ConstManager.FLOAT_EPSILON:
        return "1.0", ""
    elif my_output_norm ** 0.5 <= ConstManager.FLOAT_EPSILON:
        message = 'Cannot compare by Cosine Similarity. All the data is zero in ' + args.my_output_dump_file + '.'
        log.print_warn_log(message)
        return ConstManager.NAN, message
    elif ground_truth_norm ** 0.5 <= ConstManager.FLOAT_EPSILON:
        message = 'Cannot compare by Cosine Similarity. All the data is zero in ' + args.ground_truth_dump_file + '.'
        log.print_warn_log(message)
        return ConstManager.NAN, message

    result = ((my_output_dump_data / my_output_norm) * (ground_truth_dump_data / ground_truth_norm)).sum()
    return utils.format_value(min(result, 1.0)), ""



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import ApiAccuracyCheckerException
from msprobe.core.common.utils import is_invalid_pattern
from msprobe.mindspore.api_accuracy_checker.compute_element import ComputeElement
from msprobe.mindspore.api_accuracy_checker.utils import check_and_get_from_json_dict
from msprobe.mindspore.common.log import logger


class ApiInfo:
    def __init__(self, api_name):
        if not isinstance(api_name, str):
            err_msg = "ApiInfo.__init__ failed: api_name is not a string"
            logger.error_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed))
        if is_invalid_pattern(api_name):
            err_msg = "ApiInfo.__init__ failed: api_name contain illegal character"
            logger.error_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed))
        self.api_name = api_name
        self.forward_info = None
        self.backward_info = None

    def load_forward_info(self, forward_info_dict):
        self.forward_info = forward_info_dict

    def load_backward_info(self, backward_info_dict):
        self.backward_info = backward_info_dict

    def check_forward_info(self):
        return self.forward_info is not None

    def check_backward_info(self):
        return self.backward_info is not None

    def get_compute_element_list(self, forward_or_backward, input_or_output):
        '''
        Args:
            forward_or_backward: str, Union["forward", "backward"]
            input_or_output: str,  Union["input", "output"]

        Return:
            compute_element_list: List[ComputeElement]
        '''
        mapping = {
            (Const.FORWARD, Const.INPUT): [self.forward_info, Const.INPUT_ARGS,
                                           f"input_args field of {self.api_name} forward api in api_info.json"],
            (Const.FORWARD, Const.OUTPUT): [self.forward_info, Const.OUTPUT,
                                            f"output field of {self.api_name} forward api in api_info.json"],
            (Const.BACKWARD, Const.INPUT): [self.backward_info, Const.INPUT,
                                            f"input field of {self.api_name} backward api in api_info.json"],
            (Const.BACKWARD, Const.OUTPUT): [self.backward_info, Const.OUTPUT,
                                             f"output field of {self.api_name} backward api in api_info.json"]
        }
        dict_instance, key, key_desc = mapping.get((forward_or_backward, input_or_output))
        compute_element_info_list = check_and_get_from_json_dict(dict_instance, key, key_desc, accepted_type=list)
        compute_element_list = [ComputeElement(compute_element_info=compute_element_info)
                                for compute_element_info in compute_element_info_list]
        return compute_element_list

    def get_kwargs(self):
        '''
        Return:
            kwargs_compute_element_dict: dict{str: ComputeElement}
        '''
        kwargs_dict = check_and_get_from_json_dict(self.forward_info, Const.INPUT_KWARGS,
                                                   "input_kwargs in api_info.json", accepted_type=dict)
        for key_str, compute_element_info in kwargs_dict.items():
            if not isinstance(key_str, str):
                err_msg = "ApiInfo.get_kwargs failed: compute_element_dict key is not a string"
                logger.error_log_with_exp(err_msg,
                                          ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed))
            if not (isinstance(compute_element_info, (list, dict)) or compute_element_info is None):
                err_msg = "ApiInfo.get_kwargs failed: compute_element_dict value is not a list, dict or null"
                logger.error_log_with_exp(err_msg,
                                          ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed))
        kwargs_compute_element_dict = {key_str: ComputeElement(compute_element_info=compute_element_info)
                                       for key_str, compute_element_info in kwargs_dict.items()}
        return kwargs_compute_element_dict



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import json
from difflib import SequenceMatcher

from typing import Union, List, Dict, Any
import pandas as pd

from msprobe.core.common.utils import check_extern_input_list
from msprobe.core.config_check.checkers.base_checker import BaseChecker
from msprobe.core.config_check.config_checker import register_checker_item
from msprobe.core.config_check.utils.utils import compare_dict, config_checking_print, update_dict, \
    process_pass_check
from msprobe.core.config_check.utils.hyperparameter_parser import ParserFactory
from msprobe.core.common.file_utils import check_file_or_directory_path, create_file_in_zip, load_json, load_yaml
from msprobe.core.common.const import Const


dirpath = os.path.dirname(__file__)
hyperparameters_path = os.path.join(dirpath, "../resource/hyperparameter.yaml")
parameter_name_mapping = load_yaml(os.path.realpath(hyperparameters_path))
hyperparameters_dict = {}


def refine_json_keys(json_dcit):
    new_dict = {}
    for key in json_dcit.keys():
        new_key = key.split(Const.SEP)[-1].replace("-", "_")
        new_dict[new_key] = key 
    return new_dict       


def to_str_if_number(value):
    if isinstance(value, (int, float)):
        return str(value)
    return value


@register_checker_item("hyperparameter")
class HyperparameterChecker(BaseChecker):
    target_name_in_zip = "hyperparameters"
    result_header = ["file_name", "bench_para", "cmp_para", "bench_value", "cmp_value", "matched_with", "level"]
    hyperparameters_file_list = ["hyperparameters_static.json", "hyperparameters_dynamic.json"]

    @staticmethod
    def pack(pack_input):
        shell_path = pack_input.shell_path
        output_zip_path = pack_input.output_zip_path

        if shell_path:
            check_extern_input_list(shell_path)

            hyperparameters = {}
            parser_factory = ParserFactory()
            for script_path in shell_path:
                if os.path.isfile(script_path):
                    check_file_or_directory_path(script_path)
                    parser = parser_factory.get_parser(os.path.splitext(script_path)[1])
                    update_dict(hyperparameters, parser.run(os.path.realpath(script_path)))
                else:
                    config_checking_print(f"Warning: Script path {script_path} is not a file.")
            if hyperparameters:
                create_file_in_zip(output_zip_path,
                                   os.path.join(HyperparameterChecker.target_name_in_zip,
                                                HyperparameterChecker.hyperparameters_file_list[0]),
                                   json.dumps(hyperparameters, indent=4))
                config_checking_print(f"add static hyperparameters args to zip")
            else:
                config_checking_print(f"Warning: Failed to extract hyperparameters from script {shell_path}")
        if hyperparameters_dict:
            create_file_in_zip(output_zip_path,
                               os.path.join(HyperparameterChecker.target_name_in_zip,
                                            HyperparameterChecker.hyperparameters_file_list[1]),
                               json.dumps(vars(hyperparameters_dict), default=lambda x: None, indent=4))
            config_checking_print(f"add dynamic hyperparameters args to zip")

    @staticmethod
    def compare(bench_dir, cmp_dir, output_path):
        all_diffs = []
        for file_name in HyperparameterChecker.hyperparameters_file_list:
            bench_model_dir = os.path.join(bench_dir, HyperparameterChecker.target_name_in_zip, file_name)
            cmp_model_dir = os.path.join(cmp_dir, HyperparameterChecker.target_name_in_zip, file_name)
            if os.path.isfile(bench_model_dir) and os.path.isfile(cmp_model_dir):
                bench_hyperparameters = load_json(bench_model_dir)
                cmp_hyperparameters = load_json(cmp_model_dir)
                all_diffs.extend(
                    HyperparameterChecker.compare_param(bench_hyperparameters, cmp_hyperparameters, file_name))
        df = pd.DataFrame(all_diffs, columns=HyperparameterChecker.result_header)
        pass_check = process_pass_check(df["level"].values)
        return HyperparameterChecker.target_name_in_zip, pass_check, df

    @staticmethod
    def compare_param(bench_params, cmp_params, file_name):
        all_diffs = []
        bench_params_refined = refine_json_keys(bench_params)
        cmp_params_refined = refine_json_keys(cmp_params)

        for bench_param_name in bench_params_refined.keys():
            matched_cmp_param_name, matched_with = HyperparameterChecker._fuzzy_match_parameter(bench_param_name,
                                                                                                cmp_params_refined)
            matched_cmp_param_name = cmp_params_refined.get(matched_cmp_param_name)
            bench_param_name = bench_params_refined.get(bench_param_name)
            bench_param_value = to_str_if_number(bench_params[bench_param_name])
            if matched_cmp_param_name:
                cmp_param_value = to_str_if_number(cmp_params[matched_cmp_param_name])
                if bench_param_value != cmp_param_value:
                    all_diffs.append(
                        [file_name, bench_param_name, matched_cmp_param_name, bench_param_value, cmp_param_value,
                         matched_with, Const.CONFIG_CHECK_ERROR])
                del cmp_params[matched_cmp_param_name]
            else:
                all_diffs.append(
                    [file_name, bench_param_name, "Only in benchmark", bench_param_value, "", "",
                     Const.CONFIG_CHECK_WARNING])
        for cmp_param_name, cmp_param_value in cmp_params.items():
            all_diffs.append(
                [file_name, "Only in comparison", cmp_param_name, "", cmp_param_value, "", Const.CONFIG_CHECK_WARNING])
        all_diffs.sort()
        return all_diffs

    @staticmethod
    def apply_patches(fmk):
        try:
            from megatron import training

            def collect_hyperparameter_wrapper(func):
                def wrapper(*args, **kwargs):
                    global hyperparameters_dict
                    result = func(*args, **kwargs)
                    if not hyperparameters_dict:
                        hyperparameters_dict = result
                    return result
                return wrapper
            training.get_args = collect_hyperparameter_wrapper(training.get_args)
        except ImportError:
            config_checking_print("No megatron find.")
        except Exception as e:
            config_checking_print(f"Patch megatron method failed, detail:{str(e)}")

    @staticmethod
    def _fuzzy_match_parameter(param_name: str, available_params: Dict[str, Any]):
        """
        Fuzzy matches a parameter name against available parameter names using predefined
        mappings and string similarity.
        """
        if param_name in available_params:
            return param_name, Const.MATCH_MODE_NAME

        canonical_name = None
        for standard_name, aliases in parameter_name_mapping.items():
            if param_name == standard_name or param_name in aliases:
                canonical_name = standard_name
                break

        if canonical_name:
            if canonical_name in available_params:
                return canonical_name, Const.MATCH_MODE_MAPPING
            for alias in parameter_name_mapping[canonical_name]:
                if alias in available_params:
                    config_checking_print(
                        f"Matched '{param_name}' to alias '{alias}' via canonical name '{canonical_name}'")
                    return alias, Const.MATCH_MODE_MAPPING

        best_match_name = None
        best_match_ratio = 0.8
        for available_param_name in available_params:
            ratio = SequenceMatcher(None, param_name.lower(), available_param_name.lower()).ratio()
            if ratio > best_match_ratio:
                best_match_ratio = ratio
                best_match_name = available_param_name

        if best_match_name:
            config_checking_print(
                f"Fuzzy matched parameter '{param_name}' to '{best_match_name}' (similarity: {best_match_ratio:.2f})")
            return best_match_name, f"{Const.MATCH_MODE_SIMILARITY}:{best_match_ratio}"

        return None, None
